from warnings import warn
import segyio
from segysak._richstr import _upgrade_txt_richstr
from segysak.tools import _get_userid, _get_datetime


def _text_fixes(text):
    # hacky fixes for software
    text = text.replace("�Cro", "    ")
    text = text.replace("\x00", " ")
    return text


def _isascii(txt):
    # really want to use b"".isascii() but this is Python 3.7+
    try:
        txt.decode("ascii")
    except UnicodeDecodeError:
        return False
    else:
        return True


def get_segy_texthead(segyfile, ext_headers=False, no_richstr=False, **segyio_kwargs):
    """Return the ebcidc

    Args:
        segyfile (str): Segy File Path
        ext_headers (bool): Return EBCIDC and extended headers in list.
            Defaults to False
        no_richstr (bool, optional): Defaults to False. If true the returned string
            will not be updated for pretty HTML printing.
        segyio_kwargs: Key word arguments to pass to segyio.open
    Returns:
        str: Returns the EBCIDC text as a formatted paragraph.
    """

    with open(segyfile, mode="rb") as f:
        f.seek(0, 0)  # Locate our position to first byte of file
        data = f.read(3200)  # Read the first 3200 byte from our position

    if _isascii(data) and ext_headers == False:
        text = data.decode("ascii")  # EBCDIC encoding
        text = _text_fixes(text)
        text = segyio.tools.wrap(text)
    elif ext_headers == False:
        text = data.decode("cp500")  # text is ebcidc
        text = _text_fixes(text)
        text = segyio.tools.wrap(text)
    else:
        segyio_kwargs["ignore_geometry"] = True
        try:  # pray that the encoding is ebcidc
            with segyio.open(segyfile, "r", **segyio_kwargs) as segyf:
                text = segyf.text[0].decode("ascii", "replace")
                text = _text_fixes(text)
                text = segyio.tools.wrap(text)
                if segyf.ext_headers and ext_headers:
                    text2 = segyf.text[1].decode("ascii", "replace")
                    text = [text, text2]
        except UnicodeDecodeError as err:
            print(err)
            print("The segy text header could not be decoded.")

    if no_richstr:
        return text
    else:
        return _upgrade_txt_richstr(text)


def put_segy_texthead(segyfile, ebcidc, line_counter=True, **segyio_kwargs):

    header = ""
    if isinstance(ebcidc, dict):
        for key in ebcidc:
            if not isinstance(key, int):
                warn(
                    "ebcidc dict contains not integer keys that will be ignored",
                    UserWarning,
                )
        for line in range(1, 41):
            if line_counter:
                lc = f"C{line:02d} "
                n = 75
                content = "{:<76}"
            else:
                lc = ""
                n = 79
                content = "{:<80}"
            try:
                test = ebcidc[line]
                if len(test) > 75:
                    warn(f"EBCIDC line {line} is too long - truncating", UserWarning)
                header = header + lc + content.format(ebcidc[line][:n])
            except KeyError:
                # line not specified in dictionary
                header = header + lc + " " * n
        header = bytes(header, "utf8")
    elif isinstance(ebcidc, bytes):
        if len(ebcidc) > 3200:
            warn("Byte EBCIDC is too large - truncating", UserWarning)
        header = ebcidc[:3200]
    elif isinstance(ebcidc, str):
        if len(ebcidc) > 3200:
            warn("String EBCIDC is too large - truncating", UserWarning)
        header = bytes(ebcidc[:3200], "utf8")
    else:
        raise ValueError("Unknown ebcidc type expect bytes, str or dict")

    segyio_kwargs["ignore_geometry"] = True
    with segyio.open(segyfile, "r+", **segyio_kwargs) as segyf:
        segyf.text[0] = header


def _clean_texthead(text_dict, n=75):
    """Reduce texthead dictionary to 75 characters per line.

    The first 4 Characters of a segy EBCIDC should have the form "C01 " which
    is then follwed by 75 ascii characters.

    Input should have interger keys. Other keys will be ignored.
    Missing keys will be filled by blank lines.

    Args:
        text_dict (dict): line no and string pairs

    Returns:
        (dict): line no and string pairs ready for ebcidc input
    """
    output = dict()
    for line in range(1, 41, 1):
        try:
            line_str = text_dict[line]
            if len(line_str) > n:
                line_str = line_str[0:n]
        except KeyError:
            line_str = ""
        output[line] = line_str
    return output


def create_default_texthead(override=None):
    """Returns a simple default textual header dictionary.

    Basic fields are auto populated and a dictionary indexing lines 1-40 can
    be passed to override keyword for adjustment. By default lines 6-34 are
    empty.

    Args:
        override (dict, optional): Overide any line . Defaults to None.

    Returns:
        (dict): Dictionary with keys 1-40 for textual header of segy file

    Example:
        >>> create_default_texthead(override={7:'Hello', 8:'World!'})
        {1: 'segysak SEG-Y Output',
        2: 'Data created by: username ',
        3: '',
        4: 'DATA FORMAT: SEG-Y;  DATE: 2019-06-09 15:14:00',
        5: 'DATA DESCRIPTION: SEG-Y format data output from segysak',
        6: '',
        7: 'Hello',
        8: 'World!',
        9: '',
        ...
        34: '',
        35: '*** BYTE LOCATION OF KEY HEADERS ***',
        36: 'CMP UTM-X 181-184, ALL COORDS X100, CMP UTM-Y 185-188',
        37: 'INLINE 189-193, XLINE 194-198, ',
        38: '',
        39: '',
        40: 'END TEXTUAL HEADER'}
    """
    user = _get_userid()
    today, time = _get_datetime()
    text_dict = {
        #      123456789012345678901234567890123456789012345678901234567890123456
        1: "segysak Python Library SEG-Y Output",
        2: f"Data created by: {user} ",
        4: f"DATA FORMAT: SEG-Y;  DATE: {today} {time}",
        5: "DATA DESCRIPTION: SEG-Y format data output from segysak using segyio",
        6: "",
        40: "END TEXTUAL HEADER",
    }
    if override is not None:
        for key, line in override.items():
            text_dict[key] = line
    return _clean_texthead(text_dict)


def trace_header_map_to_text(trace_header_map):
    """Convert a trace header map to lines of text to say where header info was
    put in a file.

    Args:
        trace_header_map (dict): Header byte locations.

    Returns:
        list: List of lines to output.
    """
    header_locs_text = [f"{key}: {value}," for key, value in trace_header_map.items()]

    lines = ["*** BYTE LOCATION OF KEY HEADERS ***"]
    line = ""
    for byte_loc in header_locs_text:
        if len(line + byte_loc) < 79:
            line += byte_loc
            line += " "
        else:
            lines.append(line)
            line = byte_loc
    lines.append(line)
    return lines



#!/usr/bin/env python3

"""
Analyze one of more ZGY files to determine how bricks are ordered on disk.
Ideally the ordering is strictly one of the 6 possible permutations.
For each possibility the program will calculate how close the file is
to the goal. It will also suggest the access ordering the applications
should use for best results.

Unfortunately the check can be rather expensive for large files.
So it might not be practical to offer this as a method applications
can call at runtime. Alternatively the code might be changed to just
analyze part of the file and hope the same ordering is used for the
entire file.
"""

import numpy as np
import os, sys, math
from collections import namedtuple
from ..api import ZgyReader
from ..test.utils import SDCredentials
from ..impl import enum as impl_enum

_statisticsType = namedtuple("Statistics", "perfect partial noncont")

def iterate_InlineCrosslineSlice(reader, lod):
    lodsize = reader.brickcount[lod]
    for ii in range(lodsize[0]):
        for jj in range(lodsize[1]):
            for kk in range(lodsize[2]):
                yield (lod, ii, jj, kk)

def iterate_SliceInlineCrossline(reader, lod):
    lodsize = reader.brickcount[lod]
    for kk in range(lodsize[2]):
        for ii in range(lodsize[0]):
            for jj in range(lodsize[1]):
                yield (lod, ii, jj, kk)

def iterate_CrosslineSliceInline(reader, lod):
    lodsize = reader.brickcount[lod]
    for jj in range(lodsize[1]):
        for kk in range(lodsize[2]):
            for ii in range(lodsize[0]):
                yield (lod, ii, jj, kk)

def iterate_CrosslineInlineSlice(reader, lod):
    lodsize = reader.brickcount[lod]
    for jj in range(lodsize[1]):
        for ii in range(lodsize[0]):
            for kk in range(lodsize[2]):
                yield (lod, ii, jj, kk)

def iterate_InlineSliceCrossline(reader, lod):
    lodsize = reader.brickcount[lod]
    for ii in range(lodsize[0]):
        for kk in range(lodsize[2]):
            for jj in range(lodsize[1]):
                yield (lod, ii, jj, kk)

def iterate_SliceCrosslineInline(reader, lod):
    lodsize = reader.brickcount[lod]
    for kk in range(lodsize[2]):
        for jj in range(lodsize[1]):
            for ii in range(lodsize[0]):
                yield (lod, ii, jj, kk)

orderings = [
    ("InlineCrosslineSlice", iterate_InlineCrosslineSlice),
    ("SliceInlineCrossline", iterate_SliceInlineCrossline),
    ("CrosslineSliceInline", iterate_CrosslineSliceInline),
    ("CrosslineInlineSlice", iterate_CrosslineInlineSlice),
    ("InlineSliceCrossline", iterate_InlineSliceCrossline),
    ("SliceCrosslineInline", iterate_SliceCrosslineInline),
]

def isContiguous(reader, work, maxhole):
    """
    A single brick-column is considered to be fully optimized if bricks
    are contiguous and partly optimized if the gap between any two bricks
    is small. It is not optimized if any gap is negative, because the
    logic that consolidates bricks might refuse to handle that case.

    Note that sum(perfect, partial, noncont) will be one less than the actual
    number of bricks since the method only deals with gaps between bricks.
    """
    perfect, partial, noncont, totalsize, waste = (0, 0, 0, 0, 0)
    prev = None
    for lod, ii, jj, kk in work:
        (brickstatus, fileoffset, constvalue, bricksize) = reader._accessor._getBrickFilePosition(ii, jj, kk, lod)
        if brickstatus in (impl_enum.BrickStatus.Compressed,
                           impl_enum.BrickStatus.Normal):
            totalsize += bricksize
            if prev is not None:
                delta = fileoffset - prev
                if delta == 0:
                    perfect += 1
                elif delta > 0 and delta <= maxhole:
                    partial += 1
                    waste += delta
                else:
                    noncont += 1
            prev = fileoffset + bricksize
    return perfect, partial, noncont, totalsize, waste

def run_one_ordering(reader, lods, maxhole, ordering):
    """
    Run isContiguous() for a range of lods and sum the results.
    Also convert the result to a named tuple.
    """
    count_perfect = 0
    count_partial = 0
    count_noncont = 0
    for lod in lods:
        work = ordering(reader, lod)
        perfect, partial, noncont, _, _ = isContiguous(reader, work, maxhole)
        count_perfect += perfect
        count_partial += partial
        count_noncont += noncont
    return _statisticsType(count_perfect, count_partial, count_noncont)

def show_one_ordering(filename, reader, maxhole, ordername, ordering):
    """
    For a single ordering, e.g. InlineCrosslineSlice, check both the full
    resolution and the low resolution bricks to see how many are contiguous.
    Bricks that are less than maxhole distant count as 50% contiguous.
    Return the result as a human readable string. Also return the score
    as a number, to make it possible to pick the access order that gives
    the best result.
    """
    lod0 = run_one_ordering(reader, [0], maxhole, ordering)
    lodN = run_one_ordering(reader, range(1, reader.nlods), maxhole, ordering)
    total = (lod0.perfect + lod0.partial + lod0.noncont +
             lodN.perfect + lodN.partial + lodN.noncont)
    score = (total -
             (lod0.partial + lodN.partial) / 2 -
             (lod0.noncont + lodN.noncont)) * (100.0 / total)
    if score < 95.0:
        overall = "Suboptimal:"
    elif lod0.partial or lod0.noncont or lodN.partial or lodN.noncont:
        overall = "Acceptable:"
    else:
        overall = "Optimized: "
    if True or lod0.partial or lod0.noncont or lodN.partial or lodN.noncont:
        fmt = ("{overall} " +
               "contig {lod0.perfect:4d} " +
               "partial {lod0.partial:4d} " +
               "noncont {lod0.noncont:4d} " +
               "lowres: " +
               "contig {lodN.perfect:4d} " +
               "partial {lodN.partial:4d} " +
               "noncont {lodN.noncont:4d} " +
               "score {score:6.2f} "
               "{name} {ordername}")
    else:
        fmt = ("{overall} " +
               " size {lod0.perfect}+{lodN.perfect} bricks {name}")
    message = fmt.format(name=filename, overall=overall, lod0=lod0, lodN=lodN, score=score, ordername=ordername)
    return score, ordername, ordering, message

def show_6_orderings(filename, reader, maxhole):
    """
    Try all 6 access orders and see which of them gives the highest number
    of contiguous bricks.
    """
    results = []
    for ordername, ordering in orderings:
        result = show_one_ordering(filename, reader, maxhole, ordername, ordering)
        results.append(result)
        print(result[3])
    best = sorted(results)[-1]
    print("Recommended access order {0} score {1:6.2f} for {2}".format(best[1], best[0], filename))
    return best[1]

def run(filename, maxhole = 2*1024*1024):
    with ZgyReader(filename, iocontext = SDCredentials()) as reader:
        show_6_orderings(filename, reader, maxhole)

if __name__ == "__main__":
    np.seterr(all='raise')
    for filename in sys.argv[1:]:
        run(filename)
    #import cProfile
    #cProfile.run('run(sys.argv[1])', sort="cumulative")

# To optimize a file using the old tools:
#   env SALMON_LOD_AUTOFLUSH=0 zgycopy -v -b 64,1024,0 IN OUT
# To un-optimize it completely use --random.
# There is currently no simple way to optimize for a different ordering.

# Copyright 2017-2021, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



#!/usr/bin/env python3

"""
Test the fancy iterator.
"""

import sys
import numpy as np

from ..api import ProgressWithDots, ZgyReader, ZgyWriter, SampleDataType
from ..test.utils import SDCredentials
from ..iterator import readall

def _test_consume(iterator):
    ncalls = 0
    nvoxels = 0
    maxconsumed = [0, 0, 0]
    for pos, count, data in iterator:
        if tuple(count) != data.shape: return
        ncalls += 1
        nvoxels += int(data.size)
        for i in range(3):
            maxconsumed[i] = max(maxconsumed[i], data.shape[i])
    return ncalls, nvoxels, tuple(maxconsumed)

def test_iterator():

    class MockReader:
        def __init__(self):
            self.readcount = 0
            self.voxelcount = 0
            self.maxread = [0, 0, 0]
        def read(self, pos, data, *, lod=0):
            if False:
                print("read", tuple(pos), tuple(data.shape), lod)
            self.readcount += 1
            self.voxelcount += data.size
            for i in range(3):
                self.maxread[i] = max(self.maxread[i], data.shape[i])
        @property
        def size(self):
            return (201, 334, 1001)
        @property
        def bricksize(self):
            return (64, 64, 64)
        @property
        def datatype(self):
            return SampleDataType.int16
        @property
        def result(self):
            return self.readcount, self.voxelcount, tuple(self.maxread)

    class CountWork:
        def __init__(self, abort = None):
            self.count = 0
            self.done = 0
            self.total = 0
            self._abort = abort
        def __call__(self, done, total):
            self.count += 1
            self.done = done
            self.total = total
            return not self._abort or self.count < self._abort
        def __str__(self):
            return "done {0}/{1} in {2} calls".format(
                self.done, self.total, self.count)

    ssize = 201 * 334 * 1001
    ssize_lod1 = (202//2) * (334//2) * (1002//2)

    # For the first set of tests, leave chunksize unset so the
    # consumer gets called once for each data read.

    # If blocksize is set it will be honored, and zero means all.
    mock = MockReader()
    tmp = readall(mock, blocksize = (0, 0, 0))
    assert mock.readcount == 0
    assert _test_consume(tmp) == (1, ssize, tuple(mock.size))
    assert mock.result        == (1, ssize, tuple(mock.size))

    # Try a reasonable blocksize
    mock = MockReader()
    tmp = readall(mock, blocksize = (128, 128, 0))
    assert _test_consume(tmp) == (6, ssize, (128, 128, 1001))
    assert mock.result        == (6, ssize, (128, 128, 1001))

    # Try a silly blocksize
    mock = MockReader()
    tmp = readall(mock, blocksize = (0, 42, 501))
    assert _test_consume(tmp) == (16, ssize, (201, 42, 501))
    assert mock.result        == (16, ssize, (201, 42, 501))

    # If blocksize is not set, should end up reading the entire file
    # because it is less than 1 GB
    mock = MockReader()
    tmp = readall(mock, maxbytes=1024*1024*1024)
    assert _test_consume(tmp) == (1, ssize, (201, 334, 1001))
    assert mock.result        == (1, ssize, (201, 334, 1001))

    # If blocksize is not set but max size only has room for 3 brick-columns
    # we will be reading more in the crossline direction.
    mock = MockReader()
    tmp = readall(mock, maxbytes = 3*2048*64*64)
    assert _test_consume(tmp) == (12, ssize, (64, 128, 1001))
    assert mock.result        == (12, ssize, (64, 128, 1001))

    # Setting the blocksize to zero means as little as possible,
    # which is one brick colums unless you want to be really inefficient.
    mock = MockReader()
    tmp = readall(mock, maxbytes = 0)
    assert _test_consume(tmp) == (24, ssize, (64, 64, 1001))
    assert mock.result        == (24, ssize, (64, 64, 1001))

    # Now test setting the chunksize, leaving blocksize unset
    # causing it to read our small data in just one read.
    mock = MockReader()
    tmp = readall(mock, chunksize = (64, 64, 64), maxbytes=1024*1024*1024)
    assert _test_consume(tmp) == (4*6*16, ssize, (64, 64, 64))
    assert mock.result        == (1, ssize, tuple(mock.size))

    # As above but set a silly chunksize.
    mock = MockReader()
    tmp = readall(mock, chunksize = (100, 42, 0), maxbytes=1024*1024*1024)
    assert _test_consume(tmp) == (24, ssize, (100, 42, 1001))
    assert mock.result        == (1, ssize, tuple(mock.size))

    # Setting both.
    mock = MockReader()
    tmp = readall(mock, blocksize = (128, 128, 0), chunksize = (64, 64, 0))
    assert _test_consume(tmp) == (24, ssize, (64, 64, 1001))
    assert mock.result        == (6, ssize, (128, 128, 1001))

    # Setting chunksize to more than blocksize has no effect.
    mock = MockReader()
    tmp = readall(mock, blocksize = (64, 64, 0), chunksize = (128, 199, 0))
    assert _test_consume(tmp) == (24, ssize, (64, 64, 1001))
    assert mock.result        == (24, ssize, (64, 64, 1001))

    # A dry run may be used to get the "total" argument to the progress
    # report without actually doing any work. Pass a dummy progress
    # that saves the total and then returns False to abort.
    # Make user that (1) the iterator is consumed, so total can be saved,
    # and (2) when the iterator is consumed it doesn't actually read.
    mock = MockReader()
    progress = CountWork(abort=1)
    tmp = readall(mock, chunksize = (100, 42, 0), progress=progress)
    assert mock.readcount == 0
    assert progress.total == 0 # Because iterator not consumed yet.
    assert _test_consume(tmp) == (0, 0, (0, 0, 0))
    assert mock.result        == (0, 0, (0, 0, 0))
    assert mock.readcount == 0
    assert progress.total == ssize

    # The same run, not dry run this time, to test the progress report.
    mock = MockReader()
    progress = CountWork()
    tmp = readall(mock, chunksize = (100, 42, 0), maxbytes=1024*1024*1024, progress=progress)
    assert _test_consume(tmp) == (24, ssize, (100, 42, 1001))
    assert mock.result        == (1, ssize, tuple(mock.size))
    #print(progress)
    assert progress.count == 26 # one initial, one final, 24 regular.
    assert progress.done == ssize
    assert progress.total == ssize

    # Decimated data with silly chunksize.
    mock = MockReader()
    tmp = readall(mock, chunksize = (100, 42, 0), lod=1)
    assert _test_consume(tmp) == (8, ssize_lod1, (100, 42, 501))
    assert mock.result        == (1, ssize_lod1, (101, 167, 501))

def copyall(srcfilename, dstfilename, *, maxbytes=128*1024*1024):
    """
    Simple test that exercises the fancy iterator.
    Sets chunksize to a silly value to stress the code more.
    This is really inefficient and triggers read/modify/write
    in the writer.
    """
    p1, p2 = (ProgressWithDots(), ProgressWithDots())
    with ZgyReader(srcfilename, iocontext = SDCredentials()) as r:
        with ZgyWriter(dstfilename, templatename=srcfilename,
                       iocontext = SDCredentials()) as w:
            alldata = readall(r, maxbytes=maxbytes, progress=p1,
                              chunksize=(100, 100, 0))
            for datastart, datasize, data in alldata:
                w.write(datastart, data)
            w.finalize(progress=p2)

if __name__ == "__main__":

    test_iterator()

    if len(sys.argv) > 1:
        copyall(sys.argv[1],
                sys.argv[2],
                maxbytes = int(sys.argv[3]) if len(sys.argv) > 3 else None)

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



""" Functions needed elsewhere in the library.
"""

from attrdict import AttrDict


class FrozenDict(AttrDict):
    """A Frozen Attribute Dictionary for protected key pair mappings in SEGY-SAK."""

    def __setitem__(self, key, value):
        raise TypeError(
            "%r object does not support item assignment" % type(self).__name__
        )

    def __delitem__(self, key):
        raise TypeError(
            "%r object does not support item deletion" % type(self).__name__
        )

    def __getattribute__(self, attribute):
        if attribute in ("clear", "update", "pop", "popitem", "setdefault"):
            raise AttributeError(
                "%r object has no attribute %r" % (type(self).__name__, attribute)
            )
        return dict.__getattribute__(self, attribute)

    def __hash__(self):
        return hash(tuple(sorted(self.items())))

    def __repr__(self):
        return "FrozenDict({contents})".format(
            contents=super(AttrDict, self).__repr__()
        )

    def fromkeys(self, S, v):
        return type(self)(dict(self).fromkeys(S, v))



#!/usr/bin/env python3

"""
Code to compute statistics.

    impl.stats.StatisticData:

        * Accumulate statistics: count, sum, sum of squares, and value range.
"""
##@package openzgy.impl.stats
#@brief Code to compute statistics.

import numpy as np

class StatisticData:
    """
    Accumulate statistics: count, sum, sum of squares, and value range.
    """
    def __init__(self, other = None):
        self._cnt = other._cnt if other else 0 # Number of added samples.
        self._inf = other._inf if other else 0 # Number of rejected samples.
        self._sum = other._sum if other else 0 # Sum of added samples.
        self._ssq = other._ssq if other else 0 # Sum-of-squares of samples.
        self._min = other._min if other else 1 # Minimum sample value.
        self._max = other._max if other else -1 # Maximum  sample value.
        # Note for min and max: These initial values are only interesting
        # if storing an empty file. As soon as real data is seen they get
        # replaced (due to min > max) and the value type may also change.

    def __repr__(self):
        return "StatisticData(cnt={_cnt}, inf={_inf}, sum={_sum}, ssq={_ssq}, min={_min}, max={_max})".format(**self.__dict__)

    def __str__(self):
        return "StatisticData({_cnt} samples)".format(**self.__dict__)

    def _add_numpy_array(self, value):
        """
        Add the data in the provided numpy array.
        Technically this function should also work for Python lists or
        even scalars since all the numpy functions called are robust
        enough to handle that. But that complication isn't really needed
        so I don't want to worry about testing it.
        TODO-Low make a simpler and more efficient version for integral types.
        TODO-Low performance boost to include a "factor" argument.
        """
        if not isinstance(value, np.ndarray):
            raise TypeError("Only numpy arrays accepted when adding statistics.")
        valid = np.isfinite(value).astype(np.bool)
        scnt = np.count_nonzero(valid)
        sinf = np.count_nonzero(valid == False)
        if sinf == 0: valid = True
        # Obscure problem: initial is required for smin, smax when
        # where is used, in spite of me knowing that there is at least
        # one valid value. But, if initial=np.inf and value is an
        # integral type this goes rather badly. Using value[0] as
        # the initial is also problematic because it might be NaN.
        # To solve this I need to compute the range as float.
        # TODO-Low: or I could use int64 for all integral types but do I
        # really gain anything? Slightly more accurate results, yes,
        # but more corner cases to test and might I theoretically
        # overflow even an int64? The old code just used double.
        # TODO-Worry the existing schema stores value range as float32,
        # which is inaccurate if we ever decide to support int32 data.
        value = value.astype(np.float64, copy=False)
        if scnt:
            ssum = np.sum(value, where=valid, dtype=np.float64)
            sssq = np.sum(np.square(value, dtype=np.float64), where=valid, dtype=np.float64)
            smin = np.amin(value, where=valid, initial=np.inf)
            smax = np.amax(value, where=valid, initial=-np.inf)
            self._cnt += int(scnt)
            self._sum += ssum
            self._ssq += sssq
            if self._min is None or self._max is None or self._min > self._max:
                self._min = smin
                self._max = smax
            else:
                self._min = min(self._min, smin)
                self._max = max(self._max, smax)
        self._inf += int(sinf)
        return self

    def _add_other(self, other):
        """
        Add more samples to an existing StatisticData.
        other is also allowed to hold negative counts,
        this will cause samples to be removed. But the
        min/max range will still be expanded.
        """
        # if other.cnt_ == 0 then cnt, sum, ssq should also be zero
        # and min, max should be uninitialized. Don't trust that.
        if other._cnt != 0:
            if self._cnt != 0:
                # already have data, so the range needs to be combined.
                self._min = min(self._min, other._min)
                self._max = max(self._max, other._max)
            else:
                # our own min/max is bogus since we don't have any samples yet.
                self._min = other._min
                self._max = other._max
            self._cnt += int(other._cnt)
            self._sum += other._sum
            self._ssq += other._ssq
        self._inf += int(other._inf)
        return self

    def _multiply_scalar(self, factor):
        """
        Multiply StatisticsData with a constant N, equivalent to creating
        a new instance and adding the old one to it N times. N can also be
        negative. The min/max range is not affected.
        """
        if factor != int(factor):
            raise TypeError("Multiplication by integers only.")
        factor = int(factor) # In case it is np.int32 now.
        self._cnt *= factor;
        self._inf *= factor;
        self._sum *= factor;
        self._ssq *= factor;
        return self

    def _equal(self, other):
        if other is None:
            return False
        elif not isinstance(other, type(self)):
            raise TypeError()
        else:
            return (self._cnt == other._cnt and
                    self._inf == other._inf and
                    self._sum == other._sum and
                    self._ssq == other._ssq and
                    self._min == other._min and
                    self._max == other._max)

    def add(self, data, factor):
        if factor == 1:
            self._add_numpy_array(data)
        else:
            tmp = StatisticData()
            tmp._add_numpy_array(data)
            if factor != 1:
                tmp._multiply_scalar(factor)
            self._add_other(tmp)

    def scale(self, slope, intercept):
        """
        Calculate the linear transform needed to convert from one range
        (typically the natural data range of the integral storage type)
        to the data range that the application wants to see.
        Then update the statistics in place so they look like the transform
        had been done on every single data point before adding it.

        The decoded value Y is given by a linear transform of the coded value X:

          Y = intercept + slope*X

        where intercept and slope are given by the coding range and the value range
        of type T (see below). The statistics of Y are then:

          SUM_Y = SUM(intercept + slope*x)
                = n*intercept + slope*SUM(x) = n*intercept + slope*SUM_X

          SSQ_Y = SUM((intercept + slope*x)^2)
                = SUM(intercept^2 + 2*intercept*slope*x + slope^2*x^2)
                = n*intercept^2 + 2*intercept*slope*SUM(x) + slope^2*SUM(x^2)
                = n*intercept^2 + 2*intercept*slope*SUM_X + slope^2*SSQ_X

          MIN_Y = MIN(intercept + slope*x)
                = intercept + slope*MIN(x)
                = intercept + slope*MIN_X

          MAX_Y = MAX(intercept + slope*x)
                = intercept + slope*MAX(x)
                = intercept + slope*MAX_X
        """
        #slope = (newmax - newmin) / (oldmax - oldmin)
        #intercept = newmin - oldmin * slope
        self._ssq = self._cnt*intercept*intercept + 2*intercept*slope*self._sum + slope*slope*self._ssq;
        self._sum = self._cnt*intercept + slope*self._sum;
        self._min = intercept + slope*self._min;
        self._max = intercept + slope*self._max;
        return self

    def __eq__(self, other):
        return self._equal(other)

    def __ne__(self, other):
        return not self._equal(other)

    def __add__(self, other):
        if isinstance(other, StatisticData):
            return StatisticData(self)._add_other(other)
        elif isinstance(other, np.ndarray):
            return StatisticData(self)._add_numpy_array(other)
        else:
            raise TypeError("Can only add numpy arrays and other StatisticData instances")

    def __iadd__(self, other):
        if isinstance(other, StatisticData):
            return self._add_other(other)
        elif isinstance(other, np.ndarray):
            return self._add_numpy_array(other)
        else:
            raise TypeError("Can only add numpy arrays and other StatisticData instances")

    def __mul__(self, factor):
        return StatisticData(self)._multiply_scalar(factor)

    def __rmul__(self, factor):
        return StatisticData(self)._multiply_scalar(factor)

    def __imul__(self, factor):
        return self._multiply_scalar(factor)

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



#!/usr/bin/env python3

"""
Meta data read/write

This file contains a number of classes and free functions dealing with
meta data access, i.e. everything in the ZGY file except the actual
bulk data. Nothing in this file should be directly visible to users of
the public API.

    impl.meta.ZgyInternalMeta:

        * Holds a collection of header instances, one of each type.
        * A function to populate all the instances, mainly be invoking
          read() in each header instance, in the right order.
          With the current format there are several chicken and egg problems.
        * A function to populate all the header instances when creating
          a new file.
        * A function to flush all headers instances to file after writing.
        * The user visible ZgyReader and ZgyWriter classes should contain
          an instance of this class, using composition not inheritance
          because there are no methods or data in this class that the user
          is allowed to use directly.

    FileHeader
    OffsetHeader{V1,V2,V3,V4}
    InfoHeader{V1,V2,V3,V4}
    StringList{V1,V2,V3,V4}
    HistHeader{V1,V2,V3,V4}
    AlphaLUP{V1,V2,V3,V4}
    BrickLUP{V1,V2,V3,V4}:

        * All these represent a particular version of one of the headers
          that make up a complete ZGY file. Each header knows about the
          physical layout on the file. In C++ the class should be a POD
          with an exact match to what is found on file. If additional
          data members are needed they should be introduced in a derived
          class. In Python there are no POD types so specifying the
          physical layout is a bit more involved. And any derived data
          might as well be included in the same class.

    OffsetHeaderFactory, InfoHeaderFactory, etc.

        * Free functions to create an instance of this type, with the
          numerical version number passed in.

    HeaderBase, LookupTable

        * These are "convenience" base classes for sharing code;
          they are not intended to provide polymorphic behavior.
"""
##@package openzgy.impl
#@brief All implementation details are in this namespace.
##@package openzgy.impl.meta
#@brief Meta data read/write.

import struct
import sys
import numpy as np
import json
import math
import random
from enum import Enum

from ..impl.transform import acpToOcp
from ..impl import enum as impl_enum
from ..impl import file as impl_file
from ..exception import *

def _checked_read(f, offset, size, *, usagehint=impl_file.UsageHint.Header):
    data = f.xx_read(offset, size, usagehint=usagehint)
    if len(data) != size:
        if len(data) < size:
            raise ZgyFormatError("Got EOF reading header.")
        else:
            raise ZgyInternalError("Got too much data reading header.")
    return data

def _fix_codingrange(r, datatype):
    """
    Sanity check. If the codingrange for an int cube is bad, silently
    use a range that causes no conversion between storage and float.

    This avoids several corner cases both inside OpenZGY and in applications.

    Rationale: A non-finite range is always bad. A range with min==max
    is technically valid when reading, as all storage values would map
    to the same float value. But this is almost certainly not what the
    writer intended. Similarly a range with min>max technically means
    that increasing storage values correspond to decreasing float values.
    Again, the values are more likely to be completely bogus.

    Leave the range alone if this is a float cube. There is no conversion
    involved, and for float the codingrange is ignored by the API anyway.

    For files written by the OpenZGY library an exception wouls be thrown
    on create. So the codingrange should always be ok for those.

    Note: The sanity check could also be applied to the histogram range.
    That fix would also apply to float cubes. The histogram is less
    important though, and it should be ok to let the application worry
    about a bad histogram.
    """
    file_dtype = impl_enum._map_DataTypeToNumpyType(datatype)
    if np.issubdtype(file_dtype, np.integer):
        iinfo = np.iinfo(file_dtype)
        if (not r or
            not np.isfinite(r[0]) or
            not np.isfinite(r[1]) or
            r[1] <= r[0]):
            #print("Bad codingrange", r, "-> use", (iinfo.min, iinfo.max))
            r = (float(iinfo.min), float(iinfo.max))
    return r

class ErrorsWillCorruptFile:
    """
    Duplicated between impl.bulk and impl.meta. Maybe fix sometime.
    """
    def __init__(self, parent): self._parent = parent
    def __enter__(self): return None
    def __exit__(self, type, value, traceback):
        if type:
            #print("Meta: Exit critical section", str(value))
            self._parent._is_bad = True

class HeaderBase:
    """
    Convenience base class for implementing classes that map 1:1 to a
    specific header with a specific version that exists in a ZGY file.
    The constructor should unpack a supplied byte buffer into a new
    instance of this header. Caller is responsible for all I/O, so no
    methods in this class need an iocontext except read(). The various
    headers don't need to inherit this base class if they don't want to.
    """

    @classmethod
    def _formats(cls):
        """
        Define the physical header layout. This method needs to be
        overridden in a subclass. Otherwise the header will be empty.
        The second field is the format as recognized by the 'struct' module.
        Data on the file is stored packed and little-endian. So for use with
        'struct', a '<' should be prepended to the string.
        """
        return []

    @classmethod
    def _format(cls):
        """
        Describe the layout of this header block on the file,
        showing only the offsets as a string recognized by 'struct'.
        May be used as an argument to struct.calcsize(). Can technically
        also be used with struct.unpack to read the entire struct at
        once. But, that will returns a flat list which is tricky to
        assign to the respective attributes.
        """
        return ("<" + " ".join([ e[1] for e in cls._formats() ])).strip()

    @classmethod
    def headersize(cls):
        """
        Return the size this header has on disk.
        """
        return struct.calcsize(cls._format())

    @classmethod
    def checkformats(cls, verbose = False, *, file = None):
        """
        Helper to compare the python definition of the header layout with
        the C++ version. Also check that the same attribute isn't listed twice.
        """
        file = file or sys.stdout
        mapping = {
            "char*":   "",
            "enum":    "B",
            "float32": "f",
            "float64": "d",
            "int32":   "i",
            "int64":   "q",
            "uint32":  "I",
            "uint64":  "Q",
            "uint8":   "B",
        }
        errors = 0
        seen = set()
        byteoffset = 0
        for e in cls._formats():
            if e[0] in seen:
                print("# ERROR: attribute {0} is listed twice.".format(e[0]), file=file)
            seen.add(e[0])
            ctype = e[2]
            cname = e[0]
            csize = None
            p1 = ctype.find("[")
            p2 = ctype.find("]")
            if p1 > 0 and p2 > p1:
                csize = ctype[p1+1:p2]
                cname = cname + "[" + csize + "]"
                ctype = ctype[0:p1]
            expect = (csize if csize else '') + mapping[ctype]
            if expect == "16B": expect = "16s"
            if expect == "4B": expect = "4s"
            actual = e[1]
            if actual and actual != expect:
                print("# ERROR: Expected code {0}, got {1}".format(expect, e[1]), file=file)
                errors += 1
            byteoffset += struct.calcsize(e[1])
        assert not errors

    def pack(self):
        """
        Convert the contents of this class to a byte array suitable for
        storing in the ZGY file.
        """
        self.checkformats()
        buf = bytearray(self.headersize())
        offset = 0
        for e in self._formats():
            if e[1]:
                value = getattr(self, e[0])
                if isinstance(value, Enum):
                    value = value.value
                if isinstance(value, (list, np.ndarray)):
                    value = tuple(value)
                if not type(value) is tuple: value = (value,)
                data = struct.pack_into("<" + e[1], buf, offset, *value)
                offset += struct.calcsize("<" + e[1])
        assert offset == self.headersize()
        return buf

    def unpack(self, buf = None):
        """
        Convert a byte array as read from the ZGY file into a Python object.
        Normally a call to unpack() will immediately be followed by a call
        to calculate() to fill in any derived information, convert enume, etc.
        If buf is None, unpacking is done on an all zero buffer. This ensures
        that all data fields are present in the object. Simplifying things
        if the application is creating an instance from scratch.
        """
        self.checkformats()
        if not buf: buf = bytes(self.headersize())
        offset = 0
        for e in self._formats():
            if e[1]:
                data = struct.unpack_from("<" + e[1], buf, offset=offset)
                offset += struct.calcsize("<" + e[1])
                setattr(self, e[0], data if len(data) > 1 else data[0])
            else:
                setattr(self, e[0], None)
        assert offset == self.headersize()

    @classmethod
    def read(cls, f, offset):
        """
        Read the header from disk and parse it, returning a new instance.
        """
        return cls(_checked_read(f, offset, cls.headersize()))

    def dump(self, prefix = "", file = None):
        """
        Print the entire contents of the object, including derived fields.
        """
        file = file or sys.stdout
        print("\n{0} takes up {1} bytes with format '{2}'".format(
            self.__class__.__name__, self.headersize(), self._format()), file=file)
        for e in self._formats():
            print("{0}{1:12} = {2}".format(prefix, e[0], getattr(self, e[0])), file=file)

#****************************************************************************
#** FileHeader **************************************************************
#****************************************************************************

class FileHeader(HeaderBase):
    """
    Unpack a byte buffer into a new instance of this header.
    Caller is responsible for all I/O, so we don't need an iocontext.
    """

    def __init__(self, buf = None, *, compressed = None):
        assert self.headersize() == 8
        # Python isn't really good at overloaded methods.
        assert (buf is None and compressed is not None) or (buf is not None and compressed is None)
        if buf:
            self.unpack(buf)
            assert buf == self.pack()
            if self._magic == b'VCS\x00': raise ZgyFormatError("Old ZGY compressed files are not supported")
            if self._magic != b'VBS\x00': raise ZgyFormatError("Not an uncompressed ZGY file")
            if self._version < 1 or self._version > 4: raise ZgyFormatError("Unsupported ZGY version " + str(version))
        else:
            # Prepare for writing a new file.
            self._magic = b'VBS\x00'
            self._version = 4 if compressed else 3

    @staticmethod
    def _formats():
        return [
            ('_magic', '4s', 'uint8[4]', 'Always VBS\\0.'),
            ('_version', 'I', 'uint32', 'Current version is 3 or 4.'),
        ]

#****************************************************************************
#** OffsetHeader ************************************************************
#****************************************************************************

class OffsetHeaderV1:
    def __init__(self):
        """
        Unpack a byte buffer into a new instance of this header.
        Caller is responsible for all I/O, so we don't need an iocontext.
        """
        self._infsize = 0
        self._strsize = 0
        self._histsize = 0
        self._alphalupsize = 0
        self._bricklupsize =0
        self._infoff = -1
        self._stroff = 0 # not used in V1
        self._histoff = -1
        self._alphalupoff = -1
        self._bricklupoff = -1
        self.calculate()

    @classmethod
    def _format(self):
        return "<8I"

    @classmethod
    def size(self):
        return struct.calcsize(self._format())

    @classmethod
    def load(self, buf):
        data = struct.unpack(self._format(), buf)
        r = self()
        # Weird combination of big-endian and little-endian.
        # Also beware that it was cast to signed int.
        # Negative offsets probably yield undefined behavior.
        # But that should not have been possible anyway.
        r._infoff = (data[0] << 32) + data[1]
        r._stroff = 0
        r._alphalupoff = (data[2] << 32) + data[3]
        r._bricklupoff = (data[4] << 32) + data[5]
        r._histoff = (data[6] << 32) + data[7]
        #print("###", r._infoff)
        return r

    def calculate(self, ih = None):
        """
        Offsets are stored explicitly, so this only needs to set size.
        """
        self._infsize = InfoHeaderV1.headersize()
        self._histsize = HistHeaderV1.headersize()
        self._alphalupsize = 8 * ih._alphaoffsets[-1] if ih else 0
        self._bricklupsize = 8 * ih._brickoffsets[-1] if ih else 0

    @classmethod
    def read(cls, f):
        # The OffsetHeader always starts immediately after the FileHeader,
        # which always starts at the beginning of the file.
        return cls.load(_checked_read(f, FileHeader.headersize(), cls.size()))

    def dump(self, *, file=None):
        file = file or sys.stdout
        print("\n{0} takes up {1} bytes with format '{2}'".format(self.__class__.__name__, self.size(), self._format()), file=file)
        print("  Offsets: info {0:x} str {1:x} hist {2:x} alpha {3:x} brick {4:x}".format(
            self._infoff, self._stroff, self._histoff, self._alphalupoff, self._bricklupoff), file=file)

    @staticmethod
    def _formats():
        """
        Describe the layout of this header block on the file.
        This class doesn't inherit HeaderBase and is coded by hand.
        The implementation won't use _formats but it is included
        for the benefit of some tools.
        """
        return [
            ('_infoff',      'Q', 'int64', 'InfoHeader position in file.'),
            ('_stroff',      '',  'int64', 'String table position, N/A in V1 and pulled from InfoHeader in V2.'),
            ('_alphalupoff', 'Q', 'int64', 'Alpha tile lookup table position in file.'),
            ('_bricklupoff', 'Q', 'int64', 'Brick data lookup table position in file.'),
            ('_histoff',     'Q', 'int64', 'Histogram position in file.'),
            ]

class OffsetHeaderV2:

    def __init__(self, buf = None):
        """
        Unpack a byte buffer into a new instance of this header.
        Caller is responsible for all I/O, so we don't need an iocontext.
        """
        self._infsize = 0
        self._strsize = 0
        self._histsize = 0
        self._alphalupsize = 0
        self._bricklupsize =0
        self._infoff = -1
        self._stroff = -1
        self._histoff = -1
        self._alphalupoff = -1
        self._bricklupoff = -1
        self.calculate()

    def pack(self):
        return bytes(1)

    @classmethod
    def _format(self):
        return "<B"

    @classmethod
    def size(self):
        return struct.calcsize(self._format())

    @classmethod
    def read(cls, f):
        # This will read and discard a single byte.
        # The OffsetHeader always starts immediately after the FileHeader,
        # which always starts at the beginning of the file.
        return cls(_checked_read(f, FileHeader.headersize(), cls.size()))

    def dump(self, *, file=None):
        file = file or sys.stdout
        print("\n{0} takes up {1} bytes with format '{2}'".format(self.__class__.__name__, self.size(), self._format()), file=file)
        print(("  Offsets = info {_infoff} str {_stroff} hist {_histoff} alpha {_alphalupoff} brick {_bricklupoff}\n" +
               "  Sizes   = info {_infsize} str {_strsize} hist {_histsize} alpha {_alphalupsize} brick {_bricklupsize}" +
               "").format(**self.__dict__), file=file)

    def calculate(self, ih = None):
        """
        Calculate offsets and sizes for the various headers and tables.
        Some information requires the InfoHeader to be already known.
        If it isn't we will just calculate as much as we can.

        In general the size of a header as written to file might be
        larger than the size that the header expects to unpack.
        This allows adding more data fields at the end of the header.
        Older readers will just unpack the fields they know about.

        For ZGY V{2,3,4} this is moot, as all the offsets are implicit
        with all the headers written sequentially. So the size needs
        to match exactly or the headers following this will be corrupt.
        """
        self._infoff = FileHeader.headersize() + OffsetHeaderV2.size()
        self._infsize = InfoHeaderV2.headersize()
        if ih:
            self._strsize = ih._slbufsize
            self._histsize = HistHeaderV2.headersize()
            self._alphalupsize = 8 * ih._alphaoffsets[-1]
            self._bricklupsize = 8 * ih._brickoffsets[-1]
            self._stroff = self._infoff + self._infsize
            self._histoff = self._stroff + self._strsize
            self._alphalupoff = self._histoff + self._histsize
            self._bricklupoff = self._alphalupoff + self._alphalupsize

    @staticmethod
    def _formats():
        """
        Describe the layout of this header block on the file.
        This class doesn't inherit HeaderBase and is coded by hand.
        The implementation won't use _formats but it is included
        for the benefit of some tools.
        """
        return [
            ('_infoff',      '', 'int64', 'InfoHeader position in file.'),
            ('_stroff',      '', 'int64', 'String table position, N/A in V1 and pulled from InfoHeader in V2.'),
            ('_alphalupoff', '', 'int64', 'Alpha tile lookup table position in file.'),
            ('_bricklupoff', '', 'int64', 'Brick data lookup table position in file.'),
            ('_histoff',     '', 'int64', 'Histogram position in file.'),
            ]

class OffsetHeaderV3(OffsetHeaderV2):
    pass

class OffsetHeaderV4(OffsetHeaderV3):
    pass

def OffsetHeaderFactory(version):
    try:
        return [OffsetHeaderV1, OffsetHeaderV2, OffsetHeaderV3, OffsetHeaderV4][version-1]
    except IndexError:
        raise ZgyFormatError("Version " + str(version) + " is not supported")

#****************************************************************************
#** InfoHeader **************************************************************
#****************************************************************************

class InfoHeaderV1(HeaderBase):

    def __init__(self, buf = None):
        """
        Unpack a byte buffer into a new instance of this header.
        Caller is responsible for all I/O, so we don't need an iocontext.
        """
        assert self.headersize() == 146
        self.unpack(buf)
        self.calculate()
        if buf:
            # Check the roundtrip.
            assert buf == self.pack()

    @classmethod
    def load(self, buf):
        result = self()
        data = struct.unpack(self._format(), buf)
        #print(data)
        assert len(data) == 30
        return result

    def calculate(self, sl = None, hh = None):
        # See FileVersion<1>::InfoHeader::Interpret in the old library.
        self._bricksize = (64, 64, 64)
        try:
            if type(self._datatype) == int: self._datatype = impl_enum.RawDataType(self._datatype)
        except ValueError as ex:
            raise ZgyFormatError("Invalid enumeration found in ZGY file: " + str(ex))
        try:
            if type(self._coordtype) == int: self._coordtype = impl_enum.RawCoordType(self._coordtype)
        except ValueError as ex:
            self._coordtype = impl_enum.RawCoordType.Unknown
        self._ocp_world, self._ocp_annot, self._ocp_index = (
            _CalcOrderedCorners(self._orig, self._inc, self._size,
                                self._gpiline, self._gpxline,
                                self._gpx, self._gpy))
        self._lodsizes = _CalcLodSizes(self._size, self._bricksize)
        self._nlods = len(self._lodsizes)
        self._brickoffsets = _CalcLutOffsets(self._lodsizes, False)
        self._alphaoffsets = _CalcLutOffsets(self._lodsizes, True)
        # In V1 there is just a single value range, used both for the
        # histogram and for scaling integral values to float and for
        self._file_codingrange = (hh._min, hh._max) if hh else None
        # Sanity check. If codingrange is bad, silently use a range
        # that causes no conversion between storage and float.
        self._safe_codingrange = _fix_codingrange(self._file_codingrange, self._datatype)
        self._smin = hh._min if hh else None
        self._smax = hh._max if hh else None
        # Convert the V1 "coordtype" member to the V2 equivalent.
        if self._coordtype == impl_enum.RawCoordType.Meters:
            self._hdim = impl_enum.RawHorizontalDimension.Length
            self._hunitname = 'm'
            self._hunitfactor = 1.0
        elif self._coordtype == impl_enum.RawCoordType.Feet:
            self._hdim = impl_enum.RawHorizontalDimension.Length
            self._hunitname = 'ft'
            self._hunitfactor = 0.3048
        elif self._coordtype == impl_enum.RawCoordType.ArcSec:
            self._hdim = impl_enum.RawHorizontalDimension.ArcAngle
            self._hunitname = 'arcsec'
            self._hunitfactor = 1.0
        elif self._coordtype == impl_enum.RawCoordType.ArcDeg:
            self._hdim = impl_enum.RawHorizontalDimension.ArcAngle
            self._hunitname = 'deg'
            self._hunitfactor = 3600.0
        elif self._coordtype == impl_enum.RawCoordType.ArcDegMinSec:
            # value = deg*10000 + min*100 + sec
            # Not supported, nor does it work in the old code.
            self._hdim = impl_enum.RawHorizontalDimension.Unknown
            self._hunitname = '?'
            self._hunitfactor = 1.0
        else:
            self._hdim = impl_enum.RawHorizontalDimension.Unknown
            self._hunitname = ''
            self._hunitfactor = 1.0
        # V1 had no vertical unit information.
        self._vdim = impl_enum.RawVerticalDimension.Unknown
        self._vunitname = ''
        self._vunitfactor = 1.0

    @staticmethod
    def _formats():
        """
        Describe the layout of this header block on the file.
        The second field is the format as recognized by the 'struct' module.
        Data on the file is stored packed and little-endian. So for use with
        'struct', a '<' should be prepended to the string.
        """
        return [
            ('_size',      '3i', 'int32[3]', 'Integer size in inline, crossline, vertical directions.'),
            ('_orig',      '3i', 'int32[3]', 'First inline, crossline, time/depth. Only integral values allowed.'),
            ('_inc',       '3i', 'int32[3]', 'Integer increment in inline, crossline, vertical directions.'),
            ('_incfactor', '3f', 'float32[3]', 'Unused. Write as (1,1,1), ignore on read.'),
            ('_gpiline',   '4i', 'int32[4]', 'Inline component of 4 control points.'),
            ('_gpxline',   '4i', 'int32[4]', 'Crossline component of 4 control points.'),
            ('_gpx',       '4d', 'float64[4]', 'X coordinate of 4 control points.'),
            ('_gpy',       '4d', 'float64[4]', 'Y coordinate of 4 control points.'),
            ('_datatype',  'B', 'uint8', 'Type of samples in each brick: int8 = 0, int16 = 2, float32 = 6.'),
            ('_coordtype', 'B', 'uint8', 'Coordinate type: unknown = 0, meters = 1, feet = 2, degrees*3600 = 3, degrees = 4, DMS = 5.'),

            # Derived, as they are missing in v1 but present in v2.

            ('_bricksize', '', 'int32[3]', 'Brick size. Values other than (64,64,64) will likely not work.'),
            ('_file_codingrange', '', 'float32[2]', 'Rescaling interval used if datatype is non-float. Value range hint otherwise.'),

            # Data identifiers
            ('_dataid', '', 'uint8[16]', 'Data GUID, set on file creation.'),
            ('_verid', '', 'uint8[16]', 'Data version GUIDm set each time the file is changed.'),
            ('_previd', '', 'uint8[16]', 'Previous data version GUID.'),

            # Data source
            ('_srcname', '', 'char*', 'Source name.'),
            ('_srcdesc', '', 'char*', 'Source descriotion.'),
            ('_srctype', '', 'uint8', 'Source datatype.'),

            # Extent
            ('_curorig', '', 'int32[3]', 'Zero-based origin of extent spanned by the data currently in the file. Unused?'),
            ('_cursize', '', 'int32[3]', 'Size of extent spanned by the data currently in the file. Unused?'),

            # Statistics
            ('_scnt', '', 'int64', 'Count of values used to compute statistics.'),
            ('_ssum', '', 'float64', 'Sum of all "scnt" values.'),
            ('_sssq', '', 'float64', 'Sum of squared "scnt" values.'),
            ('_smin', '', 'float32', 'Statistical (computed) minimum value.'),
            ('_smax', '', 'float32', 'Statistical (computed) maximum value.'),

            ('_srvorig', '', 'float32[3]', 'Unused?'),
            ('_srvsize', '', 'float32[3]', 'Unused?'),

            # Grid definition
            ('_gdef', '', 'uint8', 'Grid definition type. Ignored on read.'),
            ('_gazim', '', 'float64[2]', 'Unused.'),
            ('_gbinsz', '', 'float64[2]', 'Unused.'),

            # Horizontal domain
            ('_hprjsys', '', 'char*', 'Free form description of the projection coordinate system. Usually not parseable into a well known CRS.'),
            ('_hdim', '', 'uint8', 'Horizontal dimension.'),
            ('_hunitfactor', '', 'float64', 'Multiply by this factor to convert from storage units to SI units.'),
            ('_hunitname', '', 'char*', 'For annotation only. Use the factor to convert to or from SI.'),

            # Vertical domain
            ('_vdim', '', 'uint8', 'Vertical dimension.'),
            ('_vunitfactor', '', 'float64', 'Multiply by this factor to convert from storage units to SI units.'),
            ('_vunitname', '', 'char*', 'For annotation only. Use the factor to convert to or from SI.'),

            # Derived information, both in v1 and v2.
            ('_ocp_world', '', 'enum', 'Ordered corner points: ((i0,j0),(iN,j0),(i0,jM),(iN,jM)'),
            ('_lodsizes', '', 'int32[lod]', 'Size of the survey at reduced level of detail.'),
            ('_nlods', '', 'int32', 'How many levels of details. 1 means only full resolution. Currently nlods will always be just enough to make the highest LOD (i.e. lowest resolution) fit in a single brick.'),
            ('_brickoffsets', '', 'int64[lod]', 'How many entries in the lookup table to skip when dealing with level N.'),
            ('_alphaoffsets', '', 'int64[lod]', 'How many entries in the lookup table to skip when dealing with level N.'),
        ]

class InfoHeaderV2(HeaderBase):

    def __init__(self, buf = None):
        """
        Unpack a byte buffer into a new instance of this header.
        Caller is responsible for all I/O, so we don't need an iocontext.
        """
        assert self.headersize() == 337
        self.unpack(buf)
        if buf:
            # Check the roundtrip.
            if buf != self.pack():
                print("FAIL",
                      " ".join([hex(x) for x in buf]),
                      " ".join([hex(x) for x in self.pack()]),
                      sep="\n")
            assert buf == self.pack()
        # Need to do this at end, because calculate() may patch the header.
        self.calculate()

    @staticmethod
    def _cast_enum(e, cls, default):
        """
        Convert an integer read from file to the correctly typed enum.
        The numerical values of these enums have been chosen to match
        what the file contains. If the file contains an impossible
        value then return the supplied default. It is safe to call this
        method twice. If the value is already an enum then it is returned
        unchanged.
        """
        if type(e) == int:
            try:
                return cls(e)
            except ValueError:
                return default
        else:
            return e

    def calculate(self, sl = None, hh = None):
        # Convert plain ints to enums. If they are already enums then this is a no-op.
        try:
            if type(self._datatype) == int:
                self._datatype = impl_enum.RawDataType(self._datatype)
        except ValueError as ex:
            # If the data type is not recognized the file will be unuseable.
            raise ZgyFormatError("Invalid enumeration found in ZGY file: " + str(ex))
        # The other enums on the file are not important; replace corrupt
        # values with a suitable dsefault.
        self._srctype = self._cast_enum(
            self._srctype, impl_enum.RawDataType, self._datatype)
        self._gdef    = self._cast_enum(
            self._gdef, impl_enum.RawGridDefinition, impl_enum.RawGridDefinition.Unknown)
        self._hdim    = self._cast_enum(
            self._hdim, impl_enum.RawHorizontalDimension, impl_enum.RawHorizontalDimension.Unknown)
        self._vdim    = self._cast_enum(
            self._vdim, impl_enum.RawVerticalDimension, impl_enum.RawVerticalDimension.Unknown)

        # Set the 5 strings if a StringHeader is provided, else clear them.
        # Normally we get called exactly once without a StringList, followed
        # by exactly one call that has the strings.
        self._srcname = sl._list[0] if sl else None
        self._srcdesc = sl._list[1] if sl else None
        self._hprjsys = sl._list[2] if sl else None
        self._hunitname = sl._list[3] if sl else None
        self._vunitname = sl._list[4] if sl else None

        # Geometry can be specified either as 4 ordered corner points
        # (_gdef = RawGridDefinition.FourPoint) or 3 arbitrary points
        # (_gdef = RawGridDefinition.ThreePoint) where both annotation
        # and world coordinates are given. There is also .Parametric
        # which specifies azimuth and spacing, but that is no longer
        # supported. If it ever was. With FourPoint the writer is still
        # required to store the annotation coordinates just in case
        # there is some disagreement over how the corners are ordered.
        # So, we treat FourPoint the same as ThreePoint. Ignore the last
        # point (nominally this is the corner opposite the origin)
        # and treat the remaining 3 points as if they were arbitraty.
        self._ocp_world, self._ocp_annot, self._ocp_index = (
            _CalcOrderedCorners(self._orig, self._inc, self._size,
                                self._gpiline, self._gpxline,
                                self._gpx, self._gpy))
        self._lodsizes = _CalcLodSizes(self._size, self._bricksize)
        self._nlods = len(self._lodsizes) # NOT _CalcNumberOfLODs(self._size)
        self._brickoffsets = _CalcLutOffsets(self._lodsizes, False)
        self._alphaoffsets = _CalcLutOffsets(self._lodsizes, True)
        # Sanity check. If codingrange is bad, silently use a range
        # that causes no conversion between storage and float.
        self._safe_codingrange = _fix_codingrange(self._file_codingrange, self._datatype)


    def _calculate_strings(self):
        strings = [self._srcname or '',
                   self._srcdesc or '',
                   self._hprjsys or '',
                   self._hunitname or '',
                   self._vunitname or '']
        strings = '\x00'.join(strings) + '\x00'
        strings = bytes(strings, "ASCII", errors='replace')
        return strings

    def calculate_write(self):
        """
        Call this when important parts of this struct has changed,
        Note that on write it is we that compute slbufsize that gets copied
        to offsetheader and used when the string list is written.
        Not the other way aroud.
        """
        self._lodsizes = _CalcLodSizes(self._size, self._bricksize)
        self._nlods = len(self._lodsizes)
        self._brickoffsets = _CalcLutOffsets(self._lodsizes, False)
        self._alphaoffsets = _CalcLutOffsets(self._lodsizes, True)
        self._slbufsize = len(self._calculate_strings())
        # For floating point data the coding range set by the user
        # on file creation is ignored; see _init_from_scratch().
        # on write it is set to the actual min/max data range as
        # measured in the statistics. This is very useful if the
        # file is later converted to an integral type without
        # explicitly setting the range. Also in general it is bad
        # to have an attribute which might be set incorrectly by
        # the user but is 99% unused so the consequence of setting
        # it wrong is unknown.
        # Note that overriding the codingrange is also done on read,
        # see openzgy.api.ZgyMeta.datarange.
        # TODO-Low, this does not support the case where data is written
        # containing huge spikes and where the codingrange is then
        # set to the correct range the data ought to be clipped to.
        if self._datatype == impl_enum.RawDataType.Float32:
            if self._smin <= self._smax:
                self._file_codingrange = (self._smin, self._smax)
            else: # No valid samples in file, still need a valid range
                self._file_codingrange = (-1, +1)
            self._safe_codingrange = self._file_codingrange

    @staticmethod
    def _formats():
        """
        Describe the layout of this header block on the file.
        The second field is the format as recognized by the 'struct' module.
        Data on the file is stored packed and little-endian. So for use with
        'struct', a '<' should be prepended to the string.

        In the Info header, the fields bricksize, datatype, and size
        are particularly important and immutable because their values
        affect the size of other blocks on the file. Codingrange is
        also immutable when storage type is integral. Because once
        bulk has been written, changing the codingrange would change
        all the sample values. Also, dataid is immutable by definition.

        Annotation origin and increment could technically be updated
        but this doesn't make much sense. Data may have been loaded
        indexed by annotation coordinates, and in that case changing
        the annotation would invalidate the data.
        """
        return [
            # Coding
            ('_bricksize', '3i', 'int32[3]', 'Brick size. Values other than (64,64,64) will likely not work.'),
            ('_datatype', 'B', 'uint8', 'Type of samples in each brick: int8 = 0, int16 = 2, float32 = 6.'),
            ('_file_codingrange', '2f', 'float32[2]', 'If datatype is integral, this is the value range samples will be scaled to when read as float. In this case it must be specified on file creation. If datatype is float then this is the value range of the data and should be set automatically when writing the file.'),

            # Data identifiers
            ('_dataid', '16s', 'uint8[16]', 'GUID set on file creation.'),
            ('_verid', '16s', 'uint8[16]', 'GUID set each time the file is changed.'),
            ('_previd', '16s', 'uint8[16]', 'GUID before last change.'),

            # Data source
            ('_srcname', '', 'char*', 'Optional name of this data set. Rarely used.'), # In StringList[0]
            ('_srcdesc', '', 'char*', 'Optional description of this data set. Rarely used.'), # in StringList[1]
            ('_srctype', 'B', 'uint8', 'Optional datatype the samples had before being stored in this file.'),

            # Extent
            ('_orig', '3f', 'float32[3]', 'First inline, crossline, time/depth. Unlike v1 these are now floating point.'),
            ('_inc', '3f', 'float32[3]', 'Increment in inline, crossline, vertical directions.'),
            ('_size', '3i', 'int32[3]', 'Size in inline, crossline, vertical directions.'),
            ('_curorig', '3i', 'int32[3]', 'Unused. Set to (0,0,0) on write and ignore on read.'),
            ('_cursize', '3i', 'int32[3]', 'Unused. Set to size on write and ignore on read.'),

            # Statistics. Not set directly from user code.
            ('_scnt', 'q', 'int64', 'Count of values used to compute statistics.'),
            ('_ssum', 'd', 'float64', 'Sum of all "scnt" values.'),
            ('_sssq', 'd', 'float64', 'Sum of squared "scnt" values.'),
            ('_smin', 'f', 'float32', 'Statistical (computed) minimum value.'),
            ('_smax', 'f', 'float32', 'Statistical (computed) maximum value.'),

            # Survey extent. Ignore on read.
            ('_srvorig', '3f', 'float32[3]', 'Unused. Set equal to orig on write. Ignore on read.'),
            ('_srvsize', '3f', 'float32[3]', 'Unused. Set to inc*size on write. Ignore on read.'),

            # Grid definition
            ('_gdef', 'B', 'uint8', 'Grid definition type. Set to 3 (enum: "FourPoint") on write. Ignored on read. See notes for a longer explanation.'),
            ('_gazim', '2d', 'float64[2]', 'Unused.'),
            ('_gbinsz', '2d', 'float64[2]', 'Unused.'),
            ('_gpiline', '4f', 'float32[4]', 'Inline component of 4 control points.'),
            ('_gpxline', '4f', 'float32[4]', 'Crossline component of 4 control points.'),
            ('_gpx', '4d', 'float64[4]', 'X coordinate of 4 control points.'),
            ('_gpy', '4d', 'float64[4]', 'Y coordinate of 4 control points.'),

            # Horizontal domain
            ('_hprjsys', '', 'char*', 'Free form description of the projection coordinate system. Usually not parseable into a well known CRS.'), # in StringList[2]
            ('_hdim', 'B', 'uint8', 'Horizontal dimension. Unknown = 0, Length = 1, ArcAngle = 2. Few applications support ArcAngle.'),
            ('_hunitfactor', 'd', 'float64', 'Multiply by this factor to convert from storage units to SI units. Applies to gpx, gpy.'),
            ('_hunitname', '', 'char*', 'For annotation only. Use hunitfactor, not the name, to convert to or from SI.'), # in StringList[3]

            # Vertical domain
            ('_vdim', 'B', 'uint8', 'Vertical dimension. Unknown = 0, Depth = 1, SeismicTWT = 1, SeismicOWT = 3.'),
            ('_vunitfactor', 'd', 'float64', 'Multiply by this factor to convert from storage units to SI units. Applies to orig[2], inc[2].'),
            ('_vunitname', '', 'char*', 'For annotation only. Use vunitfactor, not the name, to convert to or from SI.'), # in StringList[4]

            # Miscellaneous. Not set directly from user code.
            ('_slbufsize', 'I', 'uint32', 'Size of the StringList section.'),

            # Derived information, not stored in the file.
            ('_ocp_world', '', 'enum', 'Ordered corner points: ((i0,j0),(iN,j0),(i0,jM),(iN,jM)'),
            ('_lodsizes', '', 'int32[lod]', 'Size of the survey at reduced level of detail.'),
            ('_nlods', '', 'int32', 'How many levels of details. 1 means only full resolution. Currently nlods will always be just enough to make the highest LOD (i.e. lowest resolution) fit in a single brick.'),
            ('_brickoffsets', '', 'int64[lod]', 'How many entries in the lookup table to skip when dealing with level N.'),
            ('_alphaoffsets', '', 'int64[lod]', 'How many entries in the lookup table to skip when dealing with level N.'),
        ]

class InfoHeaderV3(InfoHeaderV2):
    pass

class InfoHeaderV4(InfoHeaderV3):
    pass

def InfoHeaderFactory(version):
    try:
        return [InfoHeaderV1, InfoHeaderV2, InfoHeaderV3, InfoHeaderV4][version-1]
    except IndexError:
        raise ZgyFormatError("Version " + str(version) + " is not supported")

#****************************************************************************
#** StringList **************************************************************
#****************************************************************************

class StringListV1:
    """
    ZGY V1 had no string list, so this section is empty.
    """
    def __init__(self, di):
        pass

    @classmethod
    def _format(cls):
        return ""

    @classmethod
    def size(cls, oh = None, ih = None):
        return 0

    @classmethod
    def load(cls, di, buf):
        return cls()

    @classmethod
    def read(cls, f, oh, ih):
        return cls(None)

    def dump(self, *, file=None):
        file = file or sys.stdout
        print("\n{0} takes up {1} bytes with format '{2}'".format(self.__class__.__name__, self.size(), self._format()), file=file)

class StringListV2:
    """
    The string list holds 5 null terminated strings:
    srcname, srcdesc, hprjsys, hunit.name, vunit.name.
    The reader will ignore trailing garbage. This is
    the one place we might add additional information
    without breaking existing readers.
    """
    def __init__(self, buf = None, oh = None, ih = None):
        """
        Unpack a byte buffer into a new instance of this header.
        Caller is responsible for all I/O, so we don't need an iocontext.
        """
        self._oh = oh
        self._ih = ih
        if buf:
            buf = buf.split(b'\x00')
            self._list = list([e.decode("ASCII", errors='replace') for e in buf])
        else:
            self._list = ['','','','','']

    @classmethod
    def _format(cls):
        return ""

    @classmethod
    def size(cls, oh = None, ih = None):
        """
        On read the size is stored in the offsetheader,
        which may have gotten it from the infoheader.
        On write the size depends on the strings written
        and we might not be able to trust the offsetheader.
        """
        # TODO-Worry handle write, especially the case where one of the strings
        # have been updated. Or isn't this needed because we don't allow
        # updating these?
        return oh._strsize if oh else 0

    @classmethod
    def read(cls, f, oh, ih):
        if oh._strsize > 0:
            return cls(_checked_read(f, oh._stroff, oh._strsize), oh, ih)
        else:
            return cls()

    def dump(self, *, file=None):
        file = file or sys.stdout
        print("\n{0} takes up {1} bytes with format '{2}'".format(self.__class__.__name__, self.size(self._oh, self._ih), self._format()), file=file)
        print("    " + "\n    ".join(self._list), file=file)

class StringListV3(StringListV2):
    pass

class StringListV4(StringListV3):
    pass

def StringListFactory(version):
    try:
        return [StringListV1, StringListV2, StringListV3, StringListV4][version-1]
    except IndexError:
        raise ZgyFormatError("Version " + str(version) + " is not supported")

#****************************************************************************
#** HistHeader **************************************************************
#****************************************************************************

class HistHeaderV1(HeaderBase):

    def __init__(self, buf = None):
        """
        Unpack a byte buffer into a new instance of this header.
        Caller is responsible for all I/O, so we don't need an iocontext.
        """
        assert self.headersize() == 256*4 + 8
        self.unpack(buf)
        #self.calculate()
        self._cnt = np.sum(self._bin) # In V1 there is no explicit all-count.
        if buf:
            # Check the roundtrip.
            assert buf == self.pack()

    @staticmethod
    def _formats():
        """
        Describe the layout of this header block on the file.
        The second field is the format as recognized by the 'struct' module.
        Data on the file is stored packed and little-endian. So for use with
        'struct', a '<' should be prepended to the string.
        """
        return [
            ('_max', 'f', 'float32', 'Center point of first bin.'),
            ('_min', 'f', 'float32', 'Center point of last bin.'),
            ('_bin', '256I', 'uint32[256]', 'Histogram.'),
        ]

class HistHeaderV2(HeaderBase):

    def __init__(self, buf = None):
        """
        Unpack a byte buffer into a new instance of this header.
        Caller is responsible for all I/O, so we don't need an iocontext.
        """
        assert self.headersize() == 256*8 + 16
        self.unpack(buf)
        #self.calculate()
        if buf:
            # Check the roundtrip.
            assert buf == self.pack()

    @staticmethod
    def _formats():
        """
        Describe the layout of this header block on the file.
        The second field is the format as recognized by the 'struct' module.
        Data on the file is stored packed and little-endian. So for use with
        'struct', a '<' should be prepended to the string.
        """
        return [
            ('_cnt', 'q', 'int64', 'Total number of samples.'),
            ('_min', 'f', 'float32', 'Center point of first bin.'),
            ('_max', 'f', 'float32', 'Center point of last bin.'),
            ('_bin', '256q', 'int64[256]', 'Histogram.'),
        ]

class HistHeaderV3(HistHeaderV2):
    pass

class HistHeaderV4(HistHeaderV3):
    pass

def HistHeaderFactory(version):
    try:
        return [HistHeaderV1, HistHeaderV2, HistHeaderV3, HistHeaderV4][version-1]
    except IndexError:
        raise ZgyFormatError("Version " + str(version) + " is not supported")

#****************************************************************************
#** Alpha lookup table ******************************************************
#****************************************************************************

class LookupTable:
    """
    Both the Alpha lookup table and the Brick lookup table hold a 64-bit
    file offset for each tile or brick in the file. Alpha tiles are bitmaps
    used to flag dead traces, and only have (i,j) coordinates. Bricks
    contain the actual samples and are indexed with (i, j, k).

    The size of the lookup tables depend on the survey size. The first
    entry in the lookup table is for the brick or tile (always just one)
    holding the lowest resolution. This is immediately followed by one
    or more entries for the bricks or tiles at level of detail N-1, and
    so on until the entries for lod 0. Within one lod level the first
    entry is for the lowest numbered i,j,[k]. For subsequent entries the
    i numbers vary fastest and the k (or j in the alpha case) varies
    slowest. Note that this is somewhat non intuitive as it is the
    opposite of the ordering of samples within a tile.

    In version 1 of the lookup tables the file offsets are stored in a
    somewhat quirky manner. The high 32 bits and the low 32 bits are
    both stored as little-endian integers, but the high part is stored
    first. So it is part big-endian, part little-endian.

    An offset of 0 means the corresponding brick or tile does not exist.
    An offset of 1 means the brick or tile contains all zeros and does
    not take any space on the file. An offset with the most significant
    bit set also means the brick or tile has a constant value. In this
    case the actual value is encoded in the least significant 8/16/32
    bits (depending on valuetype) of the stored offset.
    Offsets 0x8000000000000000 and 0x0000000000000001 are equivalent.
    Actually, v2 and later uses the first form while v1 used the second.
    For robustness both forms should be accepted regardless of version.
    """

    def __init__(self, buf, lupsize, mustflip):
        """
        Unpack a byte buffer into a new instance of this header.
        Caller is responsible for all I/O, so we don't need an iocontext.
        """
        self._mustflip = mustflip
        self._lupsize = lupsize
        if buf:
            self._lookup = list(struct.unpack(self._format(lupsize), buf))
            if self._mustflip: self._lookup = self._flip_array(self._lookup)
            self._lookend = self._calc_lookupsize(self._lookup, None, None)
        else:
            self._lookup = [0] * (lupsize//8)
            self._lookend = [0] * (lupsize//8)

    def pack(self):
        if self._mustflip: self._lookup = self._flip_array(self._lookup)
        result = struct.pack(self._format(len(self._lookup)*8), *self._lookup)
        if self._mustflip: self._lookup = self._flip_array(self._lookup)
        return result

    @staticmethod
    def _calc_lookupsize(lookup, eof, maxsize):
        """
        Given an index => start_offset lookup table, produce an
        index => end_offset table by assuming there are no holes
        in the allocated data.

        The function understands constant-value and compressed blocks.

        If eof and maxsize are known, the code can also make the
        following checks:

        Blocks that have a start offset > eof are unreadable and
        should be ignored. Set them to start and end at eof.
        The same applies to offsets of unknown type i.e. the most
        significant bit is 1 but the most significant byte is
        neither 0x80 (constant) nor 0xC0 (compressed).
        Blocks ending past eof should be assumed to end at eof.

        Blocks that appear to be larger than an uncompressed block are
        probably too large. This may be caused by holes in the allocated
        data. Assume the block is the same size as an uncompressed block.
        If a compressed block takes up more room than an uncompressed one
        then the writer should simply refrain from compressing it.
        But for extra robustness the code that makes use of this
        information shoukd be prepared to retry the access of the block
        really turned out to be larger.

        This method might be called unconditionally on file open, or
        called only if at least one compressed brick was found, or it
        might be deferred until the first time we read a compressed brick.

        TODO-Low: If alpha tiles are present then both brick and alpha offsets
        ought to be considered in the same pass. The way it will work now
        is that for bricks, a few bricks will appear too large because
        they are followed by some alpha tiles. This is harmless.
        For aplha tiles the end offsets will be hopelessly wrong.
        We will need to just assume 4 KB for those.
        """
        #print("@@@ _calc_lookupsize", list(map(hex, lookup[:5])))
        # make array of (offset, endoffset, type, ordinal)
        # Note, the only reason I use a structured array instead of a
        # 2d array [][4] is to get sort() to work the way I want it.
        # Is there a simpler way? Or is this just as performant?
        dtype = [('offset', np.uint64),
                 ('endpos', np.uint64),
                 ('type', np.uint64),
                 ('ordinal', np.uint64)]
        tmp = np.zeros(len(lookup), dtype=dtype)
        codeshift = np.uint64(56)
        codemask = np.uint64(0xFF) << codeshift
        tmp['offset'] = lookup
        tmp['ordinal'] = np.arange(len(lookup), dtype=np.uint64)
        tmp['type'] = tmp['offset']; tmp['type'] &= codemask
        tmp['offset'][tmp['type'] == (np.uint64(0x80) << codeshift)] = 0
        tmp['offset'][tmp['type'] == (np.uint64(0xC0) << codeshift)] &= ~(codemask)
        tmp.sort(order='offset')
        # The end of block i is the start of block i+1,
        # except the last block which ends at EOF, just use a huge number.
        tmp[:-1]['endpos'] = tmp[1:]['offset']
        tmp[-1]['endpos'] = eof or ~(np.uint64(1)<<np.uint64(63))
        tmp.sort(order='ordinal')
        # Several entries may be flagged as starting at offset 0.
        # With the above algorithm, an arbitrary one of these will
        # have its end set to the start of the first real block.
        # The rest will appear to end at zero. Get rid of this
        # odd behavior.
        tmp['endpos'][tmp['offset']==0] = 0
        if eof is not None and maxsize is not None:
            #endpos = max(offset, min(endpos, eof, offset + maxsize))
            np.minimum(tmp['endpos'], eof, out=tmp['endpos'])
            np.minimum(tmp['endpos'], tmp['offset'] + maxsize, out=tmp['endpos'])
            np.maximum(tmp['endpos'], tmp['offset'], out=tmp['endpos'])
            # This test as written also catches unrecognized block types
            # because the offsets are unsigned and all the unrecognized
            # types will have the most significant bit set.
        #print("@@@ _calc_lookupsize", tmp[:5])
        # Returning a numpy array of np.int64 is problematic because
        # _lookup is a normal Python array, and arithmetic such as
        # int(foo) - np.int64(bar) returns a Python float.
        #return tmp['endpos'].copy()
        return list(map(int, tmp['endpos']))

    @classmethod
    def _flip(cls, n):
        return ((n >> 32) & 0xFFFFFFFF) | ((n & 0xFFFFFFFF) << 32)

    @classmethod
    def _flip_array(cls, a):
        return list([cls._flip(n) for n in a])

    @classmethod
    def _format(cls, lupsize):
        fmt = "<" + str(lupsize//8) + "Q"
        assert struct.calcsize(fmt) == lupsize
        return fmt

    @classmethod
    def read(cls, f, offset, lupsize):
        return cls(_checked_read(f, offset, lupsize), lupsize)

    def dump(self, prefix = "", prefix0 = "", *, file=None):
        file = file or sys.stdout
        print("\n{0}{1} takes up {2} bytes with format '{3}'".format(
            prefix0,
            self.__class__.__name__,
            self._lupsize,
            self._format(self._lupsize)), file=file)
        print(prefix+("\n"+prefix).join(map(hex, self._lookup)), file=file)

class AlphaLUPV1(LookupTable):
    def __init__(self, buf, lupsize):
        super().__init__(buf, lupsize, mustflip = True)

class AlphaLUPV2(LookupTable):
    def __init__(self, buf, lupsize):
        super().__init__(buf, lupsize, mustflip = False)

class AlphaLUPV3(AlphaLUPV2):
    pass

class AlphaLUPV4(AlphaLUPV3):
    pass

def AlphaLUPFactory(version):
    try:
        return [AlphaLUPV1, AlphaLUPV2, AlphaLUPV3, AlphaLUPV4][version-1]
    except IndexError:
        raise ZgyFormatError("Version " + str(version) + " is not supported")

#****************************************************************************
#** Brick lookup table ******************************************************
#****************************************************************************

class BrickLUPV1(LookupTable):
    def __init__(self, buf, lupsize):
        super().__init__(buf, lupsize, mustflip = True)

class BrickLUPV2(LookupTable):
    def __init__(self, buf, lupsize):
        super().__init__(buf, lupsize, mustflip = False)

class BrickLUPV3(BrickLUPV2):
    pass

class BrickLUPV4(BrickLUPV3):
    pass

def BrickLUPFactory(version):
    try:
        return [BrickLUPV1, BrickLUPV2, BrickLUPV3, BrickLUPV4][version-1]
    except IndexError:
        raise ZgyFormatError("Version " + str(version) + " is not supported")

#****************************************************************************
#** All headers combined in a single instance *******************************
#****************************************************************************

class ZgyInternalMeta:
    """
    Holds references to all the individual headers needed to access ZGY.
    The information is stored in the instance in a format that is tightly
    coupled to the file format, so there will be one layer (but hopefully
    just one) above us that is responsible for the public API.
    """
    def __init__(self, myfile):
        self._is_bad = False
        if myfile: # when reading
            self._init_from_open_file(myfile)
        else: # when writing
            self._init_from_headers(None, None, None, None, None, None, None)

    def _assert_all_headers_allocated(self):
        """
        The following asserts might seem paranoid, but for a long time
        there was an assupption that every single access to those headers
        had to check first whether they were None. I am just proving to
        myself that all those checks can be removed.
        """
        assert self._fh is not None
        assert self._oh is not None
        assert self._ih is not None
        assert self._sl is not None
        assert self._hh is not None
        assert self._alup is not None
        assert self._blup is not None

    def _init_from_headers(self, fh, oh, ih, sl, hh, alup, blup):
        self._fh = fh
        self._oh = oh
        self._ih = ih
        self._sl = sl
        self._hh = hh
        self._alup = alup
        self._blup = blup

    def _init_from_open_file(self, f):
        """
        Read all the headers and save pointers to each of them in this instance.
        Some care is needed to read them in the correct order, as there are
        several dependencies between them. The next version of the file format
        will hopefully be a bit simpler in this respect.
        """

        # The file header contains only a magic string and the file version.
        # In V{2,3,4} the headers are stored consecutively:
        #     FileHeader OffsetHeader InfoHeader StringList HistHeader
        #     AlphaLUT BrickLUT
        # In V1 and most likely in the upcoming V4 only FileHeader and
        # OffsetHeader are known to be consecutive.
        fh = FileHeader.read(f, 0)

        # The offset header immediately follows the file header.
        # Changed in v2 and v3: Offset header is no longer used
        # (all offsets are now implicit), but due to a quirk in the
        # implementation it still occupies one byte on the file.
        # This is actually a bug waiting to happen, because that
        # byte (which is the size of a class with no members) is
        # probably compiler dependant.
        # Removing the OffsetHeader actually made the files tricker
        # to read, as the size of some sections depend on contents
        # of other sections.
        oh = OffsetHeaderFactory(fh._version).read(f)

        # 'oh' at this point may be incomplete (V{2,3,4}) but the offset
        # to the InfoHeader should be known by now. In V1 is is mostly
        # complete but is missing a few section sizes.
        ih = InfoHeaderFactory(fh._version).read(f, oh._infoff)

        # For V{2,3,4}, fill in the rest of the offsets now that
        # the InfoHeader is known.
        oh.calculate(ih)

        # Variable length strings are stored in a separate header.
        # Most (currently all) of these logically belong to InfoHeader.
        sl = StringListFactory(fh._version).read(f, oh, ih)

        hh = HistHeaderFactory(fh._version).read(f, oh._histoff)

        # For V2 and later, fill in the missing strings in InfoHeader
        # now that the StringList is known.
        # For V1 we also need to copy the value range (used for scaling)
        # from the histogram header.
        ih.calculate(sl, hh)

        alup = AlphaLUPFactory(fh._version).read(f, oh._alphalupoff, oh._alphalupsize)
        blup = BrickLUPFactory(fh._version).read(f, oh._bricklupoff, oh._bricklupsize)

        self._init_from_headers(fh, oh, ih, sl, hh, alup, blup)

    def _init_from_scratch(self, filename, *, size = None, compressed = False,
                           bricksize = None,
                           datatype = impl_enum.RawDataType.Float32,
                           datarange = None,
                           zunitdim = impl_enum.RawVerticalDimension.Unknown,
                           hunitdim = impl_enum.RawHorizontalDimension.Unknown,
                           zunitname = None, hunitname = None,
                           zunitfactor = None, hunitfactor = None,
                           zstart = 0.0, zinc = 0.0,
                           annotstart = (0, 0), annotinc = (0, 0),
                           corners = ((0,0),(0,0),(0,0),(0,0))):

        #--- Sanity checks and defaults management ---#

        if not size:
            raise ZgyUserError("size must be specified.")
        elif any([s<1 for s in size]):
            raise ZgyUserError("size must be at least 1 in each dimension.")

        if not bricksize:
            bricksize = (64, 64, 64)
        elif len(bricksize) != 3:
            raise ZgyUserError("bricksize must be specified in 3 dimensions.")
        elif any([(s<4 or not self._is_power_of_two(s)) for s in bricksize]):
            raise ZgyUserError("bricksize must be >= 4 and a power of 2.")

        # The codingrange for floating point data is special. The user is
        # not allowed to set it, and its value is not used internally.
        # To ensure that any user supplied range is really ignored we set
        # the range to NaN. In _calculate_write it will be set to the
        # statistical range before being written out. As an additional
        # bullet-proofing, to avoid surprises with older files, this
        # rule can also be enforced in api.ZgyMeta.datarange.
        # Note: For integral types I might have defaulted the datarange
        # to no conversion (-128..+127 or -32768..+32767) and also
        # silently re-ordered min and max if needed. But in both cases
        # the application is buggy. So, make a fuss.
        # A data range for integral data covering just a single value
        # (presumably the user will just write that single value to
        # the file) is also forbidden because it just isn't useful
        # and it triggers several corner cases.
        if datatype == impl_enum.RawDataType.Float32:
            datarange = (math.nan, math.nan)
        elif not datarange or len(datarange) != 2:
            raise ZgyUserError("datarange must be specified for integral types.")
        elif datarange[0] >= datarange[1]:
            raise ZgyUserError("datarange must have min < max.")
        elif not np.isfinite(datarange[0]) or not np.isfinite(datarange[1]):
            raise ZgyUserError("datarange must be finite.")

        zunitname = zunitname or ""
        hunitname = hunitname or ""
        zunitfactor = zunitfactor or 1
        hunitfactor = hunitfactor or 1

        #--- End sanity checks and defaults --#

        fh = FileHeader(buf = None, compressed = compressed) # Inits to version 3, or 4 if potentially compressed.
        oh = OffsetHeaderFactory(fh._version)() # Sets infoff, infsize only.
        ih = InfoHeaderFactory(fh._version)() # Inits to all zero

        # Fill in the parts of InfoHeader that affects headers elsewhere.
        ih._bricksize = tuple(bricksize) or (64, 64, 64)
        ih._datatype = datatype
        ih._size = tuple(size)
        ih._slbufsize = 0

        # Meta information caller is allowed to specify.
        # Naming is not 100% consistent; this is because
        # the parameters to this function wree set to match
        # the existing Python wrapper for the old ZGY.

        ih._file_codingrange = (datarange[0], datarange[1])
        ih._safe_codingrange = ih._file_codingrange
        ih._vdim = zunitdim
        ih._hdim = hunitdim
        ih._vunitname = zunitname
        ih._hunitname = hunitname
        ih._vunitfactor = zunitfactor
        ih._hunitfactor = hunitfactor
        ih._orig = (annotstart[0], annotstart[1], zstart)
        ih._inc = (annotinc[0], annotinc[1], zinc)
        ih._ocp_world = ((corners[0][0], corners[0][1]),
                         (corners[1][0], corners[1][1]),
                         (corners[2][0], corners[2][1]),
                         (corners[3][0], corners[3][1]))
        ih._gpx = (corners[0][0], corners[1][0], corners[2][0], corners[3][0])
        ih._gpy = (corners[0][1], corners[1][1], corners[2][1], corners[3][1])
        beg = (ih._orig[0], ih._orig[1])
        end = (ih._orig[0] + ih._inc[0] * (ih._size[0] - 1),
               ih._orig[1] + ih._inc[1] * (ih._size[1] - 1))
        ih._gpiline = (beg[0], end[0], beg[0], end[0])
        ih._gpxline = (beg[1], beg[1], end[1], end[1])

        # Meta information that might be updated after creation.
        # Except for dataid.
        def makeUUID():
            # See the C++ version for details.
            # TODO-Worry: Is the entropy of the random seed good enough?
            uuid = bytearray([random.randint(0,255) for i in range(16)])
            uuid[8] = (uuid[8] & 0x3f) | 0x80 # variant 1 (DCE)
            uuid[7] = (uuid[7] & 0x0f) | 0x40 # version 4 (random)
            return uuid
        ih._dataid = makeUUID()
        ih._verid = makeUUID()
        ih._previd = bytes(16)
        ih._srcname = ""
        ih._srcdesc = ""
        ih._srctype = datatype
        ih._hprjsys = ""

        ih._scnt = 0
        ih._ssum = 0.0
        ih._sssq = 0.0
        ih._smin = 0.0
        ih._smax = 0.0

        # Unused fields, required to be set this way for compatibility.
        ih._curorig = (0, 0, 0)
        ih._cursize = ih._size
        ih._srvorig = ih._orig
        ih._srvsize = (ih._size[0] * ih._inc[0], ih._size[1] * ih._inc[1], ih._size[2] * ih._inc[2])
        ih._gdef = impl_enum.RawGridDefinition.FourPoint
        ih._gazim = (0.0, 0.0)
        ih._gbinsz = (0.0, 0.0)

        # Derived information, see calculate()
        ih.calculate_write()

        # Fill in the rest of the offsets now that the InfoHeader is known.
        # Requires _slbufsize, ih._alphaoffsets, _brickoffsets
        oh.calculate(ih)

        # TODO-Low: Refactor: a more elegant way of handling this?
        sl = StringListFactory(fh._version)(ih._calculate_strings(), oh, ih)

        # Histogram gets initialized to empty
        hh = HistHeaderFactory(fh._version)()

        # Lookup tables get initialized to empty
        alup = AlphaLUPFactory(fh._version)(None, oh._alphalupsize)
        blup = BrickLUPFactory(fh._version)(None, oh._bricklupsize)

        self._init_from_headers(fh, oh, ih, sl, hh, alup, blup)

        # If "assert contig" in _flush_meta fails, this is also wrong:

        # ZGY aligns data to the basic block size, which depends on
        # data type and bricksize. This simplifies caching data.
        # If all headers are written sequentially from the start
        # of the file then it is simpler to add the padding as part
        # of the header data instead of before the first data block.
        # Forget about using leftover space in the header to store the
        # first few alpha tiles. We probably won't be writing those anyway.
        code = impl_enum._map_DataTypeToStructFormatCode(self._ih._datatype)
        bs = self._ih._bricksize
        bs = bs[0] * bs[1] * bs[2] * struct.calcsize(code)
        hdrsize = self._oh._bricklupoff + self._oh._bricklupsize
        padsize = (((hdrsize+bs-1)//bs)*bs)-hdrsize

        self._data_beg = hdrsize + padsize
        self._data_end = hdrsize + padsize

    def _flush_meta(self, f):

        # Make sure derived information is up to date and consistent.
        self._ih.calculate_write()
        self._oh.calculate(self._ih)

        # fileheader, infoheader, histheader all derive from HeaderBase
        # so they inherit pack() and unpack() from there. alup and blup
        # inherit pack() from LookupTable.
        fh   = self._fh.pack()
        oh   = self._oh.pack()
        ih   = self._ih.pack()
        sl   = self._ih._calculate_strings()
        hh   = self._hh.pack()
        alup = self._alup.pack()
        blup = self._blup.pack()

        assert len(ih) == self._oh._infsize
        assert len(sl) == self._oh._strsize
        assert len(hh) == self._oh._histsize
        assert len(alup) == self._oh._alphalupsize
        assert len(blup) == self._oh._bricklupsize

        contig = (self._oh._infoff      == len(fh) + len(oh) and
                  self._oh._stroff      == self._oh._infoff + len(ih) and
                  self._oh._histoff     == self._oh._stroff + len(sl) and
                  self._oh._alphalupoff == self._oh._histoff + len(hh) and
                  self._oh._bricklupoff == self._oh._alphalupoff + len(alup))

        # ZGY aligns data to the basic block size, which depends on
        # data type and bricksize. This simplifies caching data.
        # If all headers are written sequentially from the start
        # of the file then it is simpler to add the padding as part
        # of the header data instead of before the first data block.
        # Forget about using leftover space in the header to store the
        # first few alpha tiles. We probably won't be writing those anyway.
        pad = bytes()
        self._first_data = None
        if contig:
            code = impl_enum._map_DataTypeToStructFormatCode(self._ih._datatype)
            bs = self._ih._bricksize
            bs = bs[0] * bs[1] * bs[2] * struct.calcsize(code)
            hdrsize = self._oh._bricklupoff + self._oh._bricklupsize
            padsize = (((hdrsize+bs-1)//bs)*bs)-hdrsize
            pad = bytes(padsize)
            self._first_data = hdrsize + padsize

        # If not contiguous then just write the blocks one at a time.
        # For V{2,3,4} this will never happen. So I won't add code
        # for that case (which I will never be able to test).
        assert contig
        with ErrorsWillCorruptFile(self):
            f.xx_write(fh + oh + ih + sl + hh + alup + blup + pad,
                       0,
                       usagehint=impl_file.UsageHint.Header)

    def dumpRaw(self, *, file=None):
        self._fh.dump(file=file)
        self._oh.dump(file=file)
        self._ih.dump('    ', file=file)
        self._sl.dump(file=file)
        self._hh.dump('    ', file=file)
        self._alup.dump('   ', file=file)
        self._blup.dump('   ', file=file)

    @staticmethod
    def _is_power_of_two(n):
        """Not efficient, but will only be called a couple of times."""
        for shift in range(0, 32):
            if n == 1<<shift:
                return True
        return False

#****************************************************************************
#** Free functions **********************************************************
#****************************************************************************

def _CalcLodSizes(size, bricksize):
    """
    Compute the size of the file in bricks, for all LOD levels
    and all 3 dimensions. Indirectly also compute the number
    of LOD levels, since by definition the last LOD level
    is the one that has size (1, 1, 1) i.e. the entire cube
    is decimated enought to fit inside a single brick.
    """
    if min(bricksize) < 1: return [(0, 0, 0)]
    size = ((size[0] + bricksize[0] - 1) // bricksize[0],
            (size[1] + bricksize[1] - 1) // bricksize[1],
            (size[2] + bricksize[2] - 1) // bricksize[2])
    result = [ size ]
    while max(result[-1]) > 1:
        size = ((size[0]+1)//2, (size[1]+1)//2, (size[2]+1)//2)
        result.append(size)
    #print("## CalcLodSize", result)
    return tuple(result)

def _CalcLutOffsets(lods, isalpha):
    """
    Compute the offset into the lookup tables by LOD level.
    Return an array of offsets indexed by LOD. Also return
    (appended to the end of the result) the lookuop table size.
    The result differs depending on whether this is the alpha
    or brick LUT.
    The highest LOD is at the start of each table so it by
    definition has offset 0. Since the highest LOD level
    just has a single brick, the second highest will start
    at offset 1.
    """
    result = []
    pos = 0
    for e in reversed(lods):
        result.append(pos)
        pos = pos + e[0]*e[1]*(1 if isalpha else e[2])
    result.reverse()
    result.append(pos)
    #print("## CalcLutOffsets", result)
    return tuple(result)

def _CalcNumberOfLODs(size):
    """
    For illustration only. Use len(_CalcLodSize(...)) instead.
    This method suffers from numerical accuracy issues.
    """
    assert False
    return math.ceil(math.log(max(size))/math.log(2)) + 1 if size and max(size) >= 1 else 0

def _CalcOrderedCorners(orig, inc, size, gpiline, gpxline, gpx, gpy):
    """
    Convert three arbitrary control points to 4 ordered corner points ordered
    first il, first xl; last il, first xl, first il, last xl, last il, last xl.
    Also return the same 4 corners in annotation- and ordinal coordinates.
    The last two are redundant with size and annot but are provided as a
    convenience for the api layer.

    If the file contains world corners but no annotation, assume the writer
    used RawGridDefinition.FourPoint but without storing the apparently
    redundant annotation corners. This is contrary to spec but we will handle
    that case for robustness.

    If the conversion fails for another reason then return None because
    in that case coordinate conversion will not be possible.
    """
    no_world = math.isclose(gpx[0], gpx[1]) and math.isclose(gpy[0], gpy[1])
    no_annot = math.isclose(gpiline[0], gpiline[1]) and math.isclose(gpxline[0], gpxline[1])
    if no_annot and not no_world:
        ocp_world = list([[gpx[i], gpy[i]] for i in range(4)])
    else:
        try:
            ocp_world = acpToOcp(orig, inc, size, gpiline, gpxline, gpx, gpy)
        except RuntimeError:
            #raise ZgyFormatError("Invalid coordinates in ZGY file.")
            ocp_world = None
            #ocp_world = list([[gpx[i], gpy[i]] for i in range(4)])
    ends = (orig[0] + inc[0] * (size[0]-1),
            orig[1] + inc[1] * (size[1]-1))
    ocp_annot = ((orig[0], orig[1]),
                 (ends[0], orig[1]),
                 (orig[0], ends[1]),
                 (ends[0], ends[1]))
    ocp_index = ((0, 0),
                 (size[0]-1, 0),
                 (0, size[1]-1),
                 (size[0]-1, size[1]-1))
    # Make immutable so it is safe to return to application code.
    if ocp_world is not None: ocp_world = tuple(map(tuple, ocp_world))
    return (ocp_world, ocp_annot, ocp_index)

def checkAllFormats(*, verbose = False, file = None):
    FileHeader.checkformats(verbose=verbose, file=file)
    InfoHeaderV1.checkformats(verbose=verbose, file=file)
    InfoHeaderV2.checkformats(verbose=verbose, file=file)
    InfoHeaderV3.checkformats(verbose=verbose, file=file)
    InfoHeaderV4.checkformats(verbose=verbose, file=file)
    HistHeaderV1.checkformats(verbose=verbose, file=file)
    HistHeaderV2.checkformats(verbose=verbose, file=file)
    HistHeaderV3.checkformats(verbose=verbose, file=file)
    HistHeaderV4.checkformats(verbose=verbose, file=file)

# For now, run the consistency checks also when the module is imported.
checkAllFormats()

if __name__ == "__main__":
    pass
    #me = InfoHeaderV1()
    #me.dump("    ")
    #print(me.pack())

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



#!/usr/bin/env python3

"""
Deal with spatial location of the cube.

The grid definition in gpiline, gpxline, gpx, and gpy defines
an affine transform from inline, crossline to world X, Y.
Older ZGY files might use several ways to specify the transform.

Since an affine transform is used it is possible to define a
coordinate system where the X and Y axes are not precisely
perpendicular to each other. It is up to the application reading the
files to decide whether to accept non-orthogonal coordinate systems.
When writing files you should think long and hard about whether you
really need to allow the user to create a non orthogonal ZGY file.
If you do, you might regret it after a few decades of supporting that
feature in all your applications that read ZGY.


"FourPoint" is what new ZGY files should use on write. This provides
the coordinates of the four corners of the survey both in ordinal
(i, j), annotation (inline, crossline), and world (X, Y) coordinates.
The order of these points is:

   First inline / first crossline,
   last inline / first crossline,
   first inline / last crossline,
   last inline / last crossline.

The API will always return the lattice as "FourPoint", even if the ZGY
file uses a different way of specifying geometry.

The ordinal cornere are not stored. They are implied to be

   (          0,           0)
   (size[0] - 1,           0)
   (          0, size[1] - 1)
   (size[0] - 1, size[1] - 1)

The annotation corners are required to be stored, even though this is
redundant with "orig" and "inc":

   (orig[0],                    orig[1])
   (orig[0]+inc[0]*(size[0]-1), orig[1])
   (orig[0],                    orig[1]+inc[1]*(size[1]-1))
   (orig[0]+inc[0]*(size[0]-1), orig[1]+inc[1]*(size[1]-1))

The world X, Y coordinates are required to be stored for all four
corners, in spite of the last corner being redundant. As long as the
transform is required to be affine, three points are sufficient.

"ThreePoint" is more general. Instead of storing the four corners,
three arbitrary points are specified with both inline, crossline and
world X, Y coordinates. The last point in the arrays is unused. As
long as the points are not colinear or duplicated, this is enough to
define the affine transform.

The "Parametric" type is not used. The original intent was that with
"Parametric", the "gazim" field would contain Inline, crossline
azimuths (a.k.a. "base angle" and "side angle", respectively) as
360-degree compass direction (0 = N, 90 = E, 180 = S, 270 = W). The
"gbinsz" field would contain Inline, crossline bin size as METERS
always (i.e. NOT dependent on hdim or hunit). To my knowledge, no ZGY
files have ever been written using these fields. So, readers should
treat both "Unknown" and "Parametric" as errors.

New files should always be written using "FourPoint" to assure that
all ZGY readers handle them correctly. When reading, always assume
"ThreePoint" on the file but convert to "FourPoint" before providing
presenting the information to the user.

This means the reader should trust that the first three points in
(gpiline, gpxline) line up with the first three points in (gpx, gpy).
The reader should not trust that those points are in fact the corners.
And the reader should ignore the last point completely. ZGY files in
the wild might have been written explicitly in "ThreePoint" mode, or
they might accidentally have been written as "FourPoint" with the
corners in the wrong order. Reading in "ThreePoint" mode handles all
those cases.
"""
##@package openzgy.impl.transform
#@brief Deal with spatial location of the cube.

def generalTransform(ax0, ay0, ax1, ay1, ax2, ay2,
                     bx0, by0, bx1, by1, bx2, by2,
                     data):
    """
    Coordinate conversion based on 3 arbitrary control points
    in both the source and the target system, i.e. no need
    to calculate the transformation matrix first.

    This is my favorite code snippet for coordinate conversion.
    Almost no ambiguity about what it does and how to use it.

    The data array is converted in place. The code tries to ensure
    that the output is floating point even though the input might
    not be. This is to prevent surprises if converting the result
    to a numpy type (where types are more strict) and forgetting
    to give an explicit type.

    See iltf2d.cpp.cpp.

    TODO-Test, since this Python code was not ported from there
    (I used lattice.py in interpretation-sandbox instead),
    some extra testing is needed to verify the code.
    """

    # Make everything relative to p0
    ax1 -= ax0; ay1 -= ay0;
    ax2 -= ax0; ay2 -= ay0;
    bx1 -= bx0; by1 -= by0;
    bx2 -= bx0; by2 -= by0;

    det = ax1*ay2 - ax2*ay1; # The determinant

    if abs(det) < 1.0e-6:
        # If this is used to interpret coords from a ZGY file
        # then caller should catch the exception and either
        # just substitute a default or raise ZgyFormatError
        raise RuntimeError("Colinear or coincident points.")

    for pos in range(len(data)):
        xq = data[pos][0] - ax0;
        yq = data[pos][1] - ay0;
        s = (xq*ay2 - ax2*yq)/det;
        t = (ax1*yq - xq*ay1)/det;
        data[pos][0] = float(bx0 + s*bx1 + t*bx2);
        data[pos][1] = float(by0 + s*by1 + t*by2);

def acpToOcp(orig, inc, size, il, xl, wx, wy):
    """
    Convert 3 arbitrary control points containing annotation- and world coords
    into 4 ordered corner points according to the Petrel Ordered Corner Points
    (OCP) definition, which corresponds to these bulk data indices:
        (          0,           0)
        (size[0] - 1,           0)
        (          0, size[1] - 1)
        (size[0] - 1, size[1] - 1)
    See PetrelOrientationHandling

    This is used to convert from a ThreePoint to a FourPoint definition.
    If the definition is already FourPoint then calling this function with
    the 3 first points should return the same result.

    See OrderedCornerPoints.cpp.
    TODO-Test, since this Python code was not ported from there
    (I used lattice.py in interpretation-sandbox instead),
    some extra testing is needed to verify the code.
    """
    last = [orig[0] + inc[0] * (size[0]-1), orig[1] + inc[1] * (size[1]-1)]
    corners = [[orig[0], orig[1]],
               [last[0], orig[1]],
               [orig[0], last[1]],
               [last[0], last[1]]]
    generalTransform(il[0], xl[0], il[1], xl[1], il[2], xl[2],
                     wx[0], wy[0], wx[1], wy[1], wx[2], wy[2],
                     corners)
    return corners

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



#!/usr/bin/env python3

"""File: cppmeta.py
This massive kludge is/was used for the Python to C++ port,
giving a starting point and hopefully reducing the number
of copy/paste errors. It probably isn't much use after the
port is done, but I might keep it in case a really big
change needs to be done to the format.
WARNING: THE GENERATED CODE IS JUST A STARTING POINT.
"""

from ..impl import meta
from collections import namedtuple

_fixedType = namedtuple("Attribute", "cname dims ptype ctype csize comment atype aread")

def fixtuple(e):
    """
    from: ('_size',    '3i', 'int32[3]', '...')
    to:   ('_size[3]', '3i', 'int32', 3, '...')
    """
    cname = e[0]
    ptype = e[1]
    ctype = e[2]
    csize = 0
    dims = ""
    p1 = ctype.find("[")
    p2 = ctype.find("]")
    if p1 > 0 and p2 > p1:
        csize = ctype[p1+1:p2]
        dims = "[" + csize + "]"
        try:
            csize = int(csize)
        except ValueError:
            csize = -1
        #cname = cname + "[" + csize + "]"
        ctype = ctype[0:p1]
    try:
        ctype = {
            "float32": "float",
            "float64": "double",
            "uint8":   "std::uint8_t",
            "uint16":  "std::uint16_t",
            "uint32":  "std::uint32_t",
            "uint64":  "std::uint64_t",
            "int8":    "std::int8_t",
            "int16":   "std::int16_t",
            "int32":   "std::int32_t",
            "int64":   "std::int64_t",
        }[ctype]
    except KeyError:
        pass
    # Also suggest a fixed-length std::array type,
    # which might work better in access functions.
    if ctype == "char*":
        atype = "std::string"
        aread = 'std::string(_pod.{0} ? _pod.{0} : "")'.format(cname)
    elif csize:
        atype = "std::array<{0},{1}>".format(ctype, csize)
        aread = "ptr_to_array<{0},{1}>(_pod.{2})".format(ctype, csize, cname)
    else:
        atype = ctype
        aread = "align(_pod." + cname + ")"
    return _fixedType(cname, dims, ptype, ctype, csize, e[3], atype, aread)

def dumpformats(cls, *, file = None):
    file = file or sys.stdout
    say = lambda *args, **kwargs: print(*args, **kwargs, file=file)
    classname = cls.__name__
    basename = "I" + (classname[:-2] if classname[-2]=="V" else classname)+"Access"

    if classname[-2:] == "V1":
        say()
        say("/" * 77)
        say("///   " + classname[:-2] + "   " + "/" * (70 - len(classname)))
        say("/" * 77)
        say("""
class {0} : public IHeaderAccess
{{
public:
  static std::shared_ptr<{0}> factory(std::uint32_t version);
  // TO""""""DO: Add pure virtual functions, more or less matching the signatures
  // not of {1} but of the latest version. Then modify older versions
  // so they match the expected api. And while you are at it, move dump()
  // in the latest version to the base class and remove the others.
public:
}};""".format(basename, classname))

    fixedformats = list([fixtuple(ee) for ee in cls._formats()]) if hasattr(cls, "_formats") else []

    # ----- Physical layout: class FooV{n}

    say("\n// Python layout is:", cls._format())
    say("// Size in storage:", cls.headersize() if hasattr(cls, "headersize") else 0, "bytes")
    say("#pragma pack(1)\nclass {0}POD\n{{".format(classname))
    say("public:")
    for e in fixedformats:
        say("  {3}{0:17} {1:15} // {2}".format(e.ctype, e.cname + e.dims + ";", e.comment, ("" if e.ptype else "//")))
    say("};\n#pragma pack()")

    # ------ Read accessors class FooV{n}Access: aggregates FooV{n}
    #        and inherits FooBase. Note that no example FooBase
    #        is generated. Use the latest FooV{n} and replace all
    #        overridden methods with pure virtual methods.

    say("\nclass {0}Access : public {1}\n{{\npublic:\n  {0} _pod;".format(classname, basename))
    say("  virtual void read(const std::shared_ptr<FileADT> file, std::int64_t offset) override;")
    say("  virtual void byteswap() override;")
    say("  virtual void rawdump(std::ostream& out, const std::string& prefix = \"\") override;")
    say("  virtual void dump(std::ostream& out, const std::string& prefix = \"\") override;")
    say("public:")
    notimpl = 'throw OpenZGY::Errors::ZgyInternalError("Not implemented");'
    for e in fixedformats:
        sig1 ="virtual {0}".format(e.atype)
        sig ="  {0:34} {1}() const override".format(sig1, e.cname[1:])
        if e.ptype or e.atype == "std::string":
            say(sig + " {{ return {0.aread}; }}".format(e))
        else:
            say(sig + " { " + notimpl +  " }")
    say("};")

    # ----- FooV{n}Access::read() implementation.
    say("\nvoid\n{0}Access::read(const std::shared_ptr<FileADT>& file, std::int64_t offset, std::int64_t size)".format(classname))
    say("{\n  file->xx_read(&this->_pod, offset, sizeof(this->_pod));\n  byteswap();\n}")

    # ----- FooV{n}Access::byteswap() implementation.
    say("\nvoid\n{0}Access::byteswap()\n{{".format(classname))
    for e in fixedformats:
        if e.ptype and e.atype != "std::string" and e.ctype in (
                "float", "double",
                "std::int16_t", "std::int32_t", "std::int64_t",
                "std::uint16_t", "std::uint32_t", "std::uint64_t"):
            if e.csize == 0:
                say("  byteswapT(&_pod.{e.cname});".format(e=e))
            else:
                say("  byteswapT(&_pod.{e.cname}[0], {e.csize});".format(e=e))
        else:
            if e.ptype:
                say("  // byteswap not needed for {e.ctype} {e.cname}{e.dims} because of its type.".format(e=e))
            else:
                say("  // byteswap not needed for {e.ctype} {e.cname}{e.dims} because it is not stored.".format(e=e))
    say("}")

    # ----- Debugging: FooV{n}Access::rawdump() implementation.
    #       This uses the POD data members directly.

    say("\nvoid\n{0}Access::rawdump(std::ostream& out, const std::string& prefix)\n{{".format(classname))
    say('  out')
    for e in fixedformats:
        if e.ptype:
            say('    << prefix << "{0:20s}"'.format(e.cname+":"), end="")
        else:
            say('    //<< prefix << "{0:20s}"'.format(e.cname+":"), end="")
        if not e.csize:
            say(' << _pod.{0} << "\\n"'.format(e.cname))
        else:
            for i in range(e.csize):
                say(' << _pod.{0}[{1}] << " "'.format(e.cname, i), end="");
            say(' << "\\n"')
    say("  ;")
    say("}")

    # ----- Debugging: FooV{n}Access::dump() implementation.
    #       This uses the generated access methods.
    #       NOTE: Should probably move the highest version of
    #       FooV{n}Access::dump to the FooBase base class
    #       and remove the others.
    say("\nvoid\n{0}Access::dump(std::ostream& out, const std::string& prefix)\n{{".format(classname))
    say('  out')
    for e in fixedformats:
        say('    << prefix << "{0:20s}"'.format(e.cname[1:]+"():"), end="")
        if not e.ptype:
            say(' << "N/A\\n"')
        elif not e.csize:
            say(' << {0}() << "\\n"'.format(e.cname[1:]))
        else:
            say(' << array_to_string({0}()) << "\\n"'.format(e.cname[1:]))
    say("  ;")
    say("}")

    @classmethod
    def checkformats(cls, verbose = False, *, file = None):
        """
        This is the old class-dumper, moved from impl.meta because it is
        not production code. In fact, its functinality is superceded by
        dumpformats but might, like dumpformats, turn out to be useful
        at some point. Yagni doesn't apply here.

        Helper to compare the python definition of the header layout with
        the C++ version. If verbose is True output the entire definition
        as it would appear in a C++ header. With verbose False it only
        does the consistency check. This is cheap enough to be permanently
        enabled. Also check that the same attribute isn't listed twice.
        """
        file = file or sys.stdout
        mapping = {
            "char*":   "",
            "enum":    "B",
            "float32": "f",
            "float64": "d",
            "int32":   "i",
            "int64":   "q",
            "uint32":  "I",
            "uint64":  "Q",
            "uint8":   "B",
        }
        errors = 0
        seen = set()
        byteoffset = 0
        if verbose == 1:
            print("// Python layout is:", cls._format(), file=file)
            print("// Size in storage:", cls.headersize(), "bytes", file=file)
            print("class {0}\n{{".format(cls.__name__), file=file)
        if verbose == 2:
            print("<h3>{0}</h3>".format(cls.__name__), file=file)
            print('<table border="1" style="border-collapse: collapse">', file=file)
            print("<tr><th>offset</th><th>size</th><th>type</th><th>name</th><th>remarks</th></tr>", file=file)
        for e in cls._formats():
            if e[0] in seen:
                print("# ERROR: attribute {0} is listed twice.".format(e[0]), file=file)
            seen.add(e[0])
            ctype = e[2]
            cname = e[0]
            csize = None
            p1 = ctype.find("[")
            p2 = ctype.find("]")
            if p1 > 0 and p2 > p1:
                csize = ctype[p1+1:p2]
                cname = cname + "[" + csize + "]"
                ctype = ctype[0:p1]
            if verbose == 1:
                #print("  // offset", byteoffset, file=file)
                print("  {3}{0:10} {1:15} // {2}".format(ctype, cname + ";", e[3], ("" if e[1] else "//")), file=file)
            if verbose == 2:
                print("<tr><td>{3:3}</td><td>{4:2}</td><td>{0:10}</td><td>{1:15}</td><td>{2}</td></tr>".format(ctype, cname, e[3], byteoffset, struct.calcsize(e[1])), file=file)
            expect = (csize if csize else '') + mapping[ctype]
            if expect == "16B": expect = "16s"
            if expect == "4B": expect = "4s"
            actual = e[1]
            if actual and actual != expect:
                print("# ERROR: Expected code {0}, got {1}".format(expect, e[1]), file=file)
                errors += 1
            byteoffset += struct.calcsize(e[1])
        if verbose == 1:
            print("};", file=file)
        if verbose == 2:
            print('<tr><td>{0:3}</td><td colspan="3">&nbsp;</td><td>end</td></tr>'.format(byteoffset), file=file)
            print("</table>", file=file)
        assert not errors

def checkAllFormats(*, verbose = False, file = None):
    meta.FileHeader.checkformats(verbose=verbose, file=file)
    meta.InfoHeaderV1.checkformats(verbose=verbose, file=file)
    meta.InfoHeaderV2.checkformats(verbose=verbose, file=file)
    meta.InfoHeaderV3.checkformats(verbose=verbose, file=file)
    meta.HistHeaderV1.checkformats(verbose=verbose, file=file)
    meta.HistHeaderV2.checkformats(verbose=verbose, file=file)
    meta.HistHeaderV3.checkformats(verbose=verbose, file=file)

def dumpAllFormats(*, file = None):
    print("// AUTOGENERATED -- NEEDS MAJOR EDITS BEFORE USE\n", file=f)
    dumpformats(meta.FileHeader, file=f)
    dumpformats(meta.OffsetHeaderV1, file=f)
    dumpformats(meta.OffsetHeaderV2, file=f)
    dumpformats(meta.InfoHeaderV1, file=f)
    dumpformats(meta.InfoHeaderV2, file=f)
    dumpformats(meta.HistHeaderV1, file=f)
    dumpformats(meta.HistHeaderV2, file=f)

if __name__ == "__main__":
    # Consistency check only
    checkAllFormats()

    # Simple C++ header file (dumpformats does much more)
    #checkAllFormats(verbose=1)

    # HTML formatted documentation
    #checkAllFormats(verbose=2)

    with open("tmpmetatmp.h", "w") as f:
        dumpAllFormats(file = f)

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



#!/usr/bin/env python3

"""
Enums not visible to the public API.

Some of the enums are used to describe parts of the file format,
giving a symbolic name to an integer stored in the file.
That kind of information should definitely be hidden from clients.
"""
##@package openzgy.impl.enum
#@brief Enums not visible to the public API.

from enum import Enum
import numpy as np

class RawDataType(Enum):
    """
    Sample data type as stored on the file.
    In the public API this maps to SampleDataType.
    Source: BrickedFileVersion.cpp, MetaDataValue.h
    This enum is used for all versions.
    Note that the existing public ZGY library
    only recognizes SignedInt8, SignedInt16, and Float32.
    """
    SignedInt8    = 0
    UnsignedInt8  = 1
    SignedInt16   = 2
    UnsignedInt16 = 3
    SignedInt32   = 4
    UnsignedInt32 = 5
    Float32       = 6
    IbmFloat32    = 7

def _map_DataTypeToStructFormatCode(dt):
    """Internal method to get the "struct" format code."""
    return {
        RawDataType.SignedInt8:    "<b",
        RawDataType.UnsignedInt8:  "<B",
        RawDataType.SignedInt16:   "<h",
        RawDataType.UnsignedInt16: "<H",
        RawDataType.SignedInt32:   "<i",
        RawDataType.UnsignedInt32: "<I",
        RawDataType.Float32:       "<f",
    }[dt]

def _map_DataTypeToNumpyType(dt):
    """
    Internal method to get the "numpy" format code.
    Caveat: This returns a type object inheriting numpy.number,
    not a numpy.dtype instance. These are often interchangeable
    but not always.
    """
    return {
        RawDataType.SignedInt8:    np.int8,
        RawDataType.UnsignedInt8:  np.uint8,
        RawDataType.SignedInt16:   np.int16,
        RawDataType.UnsignedInt16: np.uint16,
        RawDataType.SignedInt32:   np.int32,
        RawDataType.UnsignedInt32: np.uint32,
        RawDataType.Float32:       np.float32,
    }[dt]

class RawCoordType(Enum):
    """
    Coordinate type codes as stored in V1 files only.
    The values are stored on the file, so the numbers must not be changed.
    Source: BrickedFileVersion.cpp.
    There is no corresponding enum in the API layer.
    """
    Unknown        = 0
    Meters         = 1
    Feet           = 2
    ArcSec         = 3  # value = deg*3600 + min*60 + sec
    ArcDeg         = 4  # value = deg + min/60 + sec/3600
    ArcDegMinSec   = 5  # value = deg*10000 + min*100 + sec

class RawHorizontalDimension(Enum):
    """
    Horizontal dimension as seen in V2 files and later.
    In the public API this maps to UnitDimension.
    The values are stored in the file, so the numbers must not be changed.
    Source: PoststackSeis3dInfo.h, MetaDataValue.h, ReaderImp::getMetaData
    """
    Unknown = 0
    Length = 1
    ArcAngle = 2

class RawVerticalDimension(Enum):
    """
    Vertical dimension as seen in V2 files and later.
    In the public API this maps to UnitDimension.
    The values are stored in the file, so the numbers must not be changed.
    Source: PoststackSeis3dInfo.h, MetaDataValue.h, ReaderImp::getMetaData
    """
    Unknown = 0
    Depth = 1
    SeismicTWT = 2
    SeismicOWT = 3

class RawGridDefinition(Enum):
    """
    Method used to define the geometry. Only FourPoint is allowed for write,
    and only ThreePoint (treated as FourPoint) and FourPoint supported on read.
    The values are stored in the file, so the numbers must not be changed.
    There is no corresponding enum in the API layer.
    """
    Unknown    = 0
    Parametric = 1
    ThreePoint = 2
    FourPoint  = 3

class BrickStatus(Enum):
    """
    Brick status as used in the internal API only.
    """
    Missing = 0
    Constant = 1
    Normal = 2
    Compressed = 3

class UpdateMode(Enum):
    """
    WORK IN PROGRESS, potential configurable behavior.

    A ZGY file cannot be updated once created, but individual bricks might
    be written to more than once while the file is still open for create.

    Updating a brick might cause loss of quality if the update was made
    as part of a read/modify/write cycle. It might also cause space to be
    wasted in the file since ZGY does not try to recycle freed bricks.
    For this reason the application should explicitly indicate that it
    accepts the loss of quality and/or leakage.

    Kinds of leakage:

     -  Brick to be overwritten is in a closed segment. This is
        expected to be rare, and only relevant for cloud storage.

     -  Brick to be overwritten and/or new brick is compressed
        and the new data is smaller. Leaks the size difference,
        although for implementation reasons we might want to leak
        the entire old brick ("Pedantic" mode).

     -  Brick to be overwritten and/or new brick is compressed
        and the new data is larger. Leaks the old brick.

    The default is "Always" for uncompressed local files and "Constant"
    otherwise.

    It is fairly safe to set an uncompressed cloud file to "Always" but
    there are some scenarios where very small regions are written to a
    large file where this might cause much leakage. So the caller
    needs to confirm he knows what he is doing.

    Compressed files should only be set to "Always" in very special cases
    or in unit tests. The potential leakage is much larger, as is the
    problem of multiple compress and decompress cycles causing noise.
    """

    Never = 0       # Never allow updating. Can only write to "Missing" bricks.

    Constant = 1    # Can write to both "Missing" and "Constant" bricks. This
                    # permission is needed if the client wants to establish
		    # a default value for missing samples by setting the entire
		    # survey to this default. Followed by writing real data.
		    # "Never" and "Constant" require that the application
		    # writes brick aligned data. If a read/modify/write is
		    # indicated then this will raise an exception.
		    # The only drawback of "Constant" over "Never" is that
		    # "Constant" can cause a slight confusion: A particular
		    # read/modify/write might be allowed if the previous write
		    # just happened to have all constant values.

    #NoCompress = 2 # Not recommended because it gives confusing behavior.
                    # Update is only allowed when both the old and the new
                    # brick is uncompressed. The only leaks allowed are those
                    # caused by the target block being in a closed segment on
                    # the cloud. The problem is that the decision is made
                    # per brick. It is more consistent to decide up front
                    # whether any compression might happen. If so, use
                    # Constant or Never. If not, just use Always.

    #NoLeaks = 3    # Not recommended because it gives confusing behavior.
                    # As "Always" but if it turns out that a brick would be
                    # leaked, even in the rare "closed segment" case, the
                    # code will raise an exception.

    Always = 4      # Always allow updating. This may cause leaked data in
                    # some cases. Uncompressed local data will not leak.
                    # Uncompressed cloud data can only leak when the target
                    # block being in a closed segment on the cloud.
		    # This can be the default for uncompressed data.
		    # For compressed data it is usually a bad idea.
                    # Not only could there be a lot more leakage, but if
                    # the reason for the overwrite is a read/modify/write
                    # and the data had lossy compression then the
                    # compression noise will accumulate.

    Pedantic = 5    # As "Always", but when either the existing or the new
                    # brick is compressed then the old brick is leaked
                    # unconditionally. This gives more reproducible behavior
                    # but wastes more space.

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



"""Enumerations of dataset keys.
"""

from ._core import FrozenDict

CoordKeyField = FrozenDict(
    dict(
        cdp="cdp",
        iline="iline",
        xline="xline",
        offset="offset",
        depth="depth",
        twt="twt",
        cdp_x="cdp_x",
        cdp_y="cdp_y",
        lat="lat",
        long="lon",
    )
)

DimensionKeyField = FrozenDict(
    dict(
        cdp_3d=("iline", "xline"),
        cdp_2d=("cdp",),
        threed_twt=("iline", "xline", "twt"),
        threed_xline_twt=("iline", "twt"),
        threed_iline_twt=("xline", "twt"),
        threed_depth=("iline", "xline", "depth"),
        threed_xline_depth=("iline", "depth"),
        threed_iline_depth=("xline", "depth"),
        threed_head=("iline", "xline"),
        threed_ps_twt=("iline", "xline", "twt", "offset"),
        threed_ps_depth=("iline", "xline", "depth", "offset"),
        threed_ps_head=("iline", "xline", "offset"),
        twod_twt=("cdp", "twt"),
        twod_depth=("cdp", "depth"),
        twod_head=("cdp",),
        twod_ps_twt=("cdp", "twt", "offset"),
        twod_ps_depth=("cdp", "depth", "offset"),
        twod_ps_head=("cdp", "offset"),
    )
)

VariableKeyField = FrozenDict(dict(data="data"))


AttrKeyField = FrozenDict(
    dict(
        ns="ns",
        sample_rate="sample_rate",
        text="text",
        measurement_system="measurement_system",
        d3_domain="d3_domain",
        epsg="epsg",
        corner_points="corner_points",
        corner_points_xy="corner_points_xy",
        source_file="source_file",
        srd="srd",
        datatype="datatype",
        percentiles="percentiles",
        coord_scalar="coord_scalar",
    )
)


VerticalKeyField = FrozenDict(dict(twt="TWT", depth="DEPTH"))
VerticalKeyDim = FrozenDict(dict(TWT="twt", DEPTH="depth"))


VerticalUnits = FrozenDict(
    dict(
        ms="ms",  # milliseconds
        s="s",  # seconds
        cm="cm",  # centimetres
        m="m",  # metres
        km="km",  # kilometres
        ft="ft",  # feet
    )
)



#!/usr/bin/env python3

"""
Implement a simple copy command that exercises the new code.
Optionally crop the input.
"""

import numpy as np
import os, sys, time, argparse
from ..api import ZgyReader, ZgyWriter, SampleDataType, ProgressWithDots, ZgyCompressFactory
from ..test.utils import SDCredentials
from ..iterator import readall

def suggest_range(value, dt):
    """
    Special case handling for inconsistent all-constant files.
    A coding range with min==max is not valid, so we must choose
    something else.
    """
    dt_lo, dt_hi = {SampleDataType.int8: (-128, +127),
                    SampleDataType.int16: (-32768, +32767)}[dt]
    if value == 0:
        # Choose the range -1..+1, but slightly wider at the low end
        # to make float-0 map exactly to int-0, zero centric.
        return (dt_lo / dt_hi, 1)
    elif value > 0:
        # Choose the range 0..2*value, but slightly less at the high end
        # to make float-value map exactly to int-0, zero centric.
        return (0, value * (1 - dt_hi / dt_lo))
    else:
        # Choose the range 2*value..0, but slightly wider at the low end
        # to make float-value map exactly to int-0, zero centric.
        return (value * (1 - dt_lo / dt_hi), 0)

def find_center(r, maxsize):
    """
    Crop partial bricks outside or below the survey since they
    would just confuse the statistics.
    If the file is large then crop out an area around the center.
    Try to read the full traces but limit the il and xl.
    """
    size = np.array(r.size, dtype=np.int64)
    cropsize = np.minimum(size, np.array(maxsize, dtype=np.int64))
    cropsize = (cropsize//64)*64
    cropoffset = ((size - cropsize)//128)*64
    if np.any(size - cropsize >= 64):
        print("Reading center", tuple(cropsize),
              "offset", tuple(cropoffset),
              "of survey size", tuple(r.size))
    return tuple(cropsize), tuple(cropoffset)

def copy_open_file(r, w, *, progress, offset=(0,0,0), noisefactor=0):
    noiselevel = abs(w.datarange[1] - w.datarange[0]) / noisefactor if noisefactor > 0 else None
    for datastart, datasize, data in readall(r, dtype=np.float32, cropsize=w.size, cropoffset=offset, progress=progress):
        if noiselevel:
            # Note, maybe multiply with a random factor as well,
            # if the goal is to test compression when white noise
            # is present but where the input used to be integral.
            data += (np.random.random_sample(view.shape) - 0.5) * noiselevel
        w.write(datastart, data)

def read_and_discard(r, cropsize, *, progress, offset=(0,0,0), noisefactor=0):
    """This is for performance testing only."""
    for datastart, datasize, data in readall(r, dtype=np.float32, cropsize=cropsize, cropoffset=offset, progress=progress):
        pass

def copy(srcfilename, dstfilename, crop_offset=None, crop_size=None, crop_center=None, forcetype='unchanged', datarange=None, noisefactor=0, snr = 0, progress1 = None, progress2 = None):
    starttime = time.time()
    with ZgyReader(srcfilename, iocontext = SDCredentials()) as r:
        if crop_center:
            if crop_offset and crop_offset != (0,0,0):
                print("WARNING: ignoring --offset because --center specified.")
            if not crop_size or crop_size == (0,0,0):
                crop_size = (640, 640, 12800)
            crop_size, crop_offset = find_center(r, crop_size)
        # Zero cropped size means no cropping, i.e. "to end of survey".
        # TODO-Low passing crop_size but not crop_offset could mean "center".
        crop_beg  = crop_offset or (0,0,0)
        crop_size = crop_size or (0,0,0)
        crop_beg = tuple([crop_beg[i] if crop_beg[i] >= 0 else r.size[i] + crop_beg[i] for i in range(3)])
        crop_size = tuple([crop_size[i] or r.size[i]-crop_beg[i] for i in range(3)])
        crop_end  = tuple([crop_beg[i] + crop_size[i] - 1 for i in range(3)])
        # Need to re-calculate the corners and the annotation
        index_corners = [ [ crop_beg[0], crop_beg[1] ],
                          [ crop_end[0], crop_beg[1] ],
                          [ crop_beg[0], crop_end[1] ],
                          [ crop_end[0], crop_end[1] ] ]
        world_corners = [ r.indexToWorld(tuple(x)) for x in index_corners ]
        annotstart = r.indexToAnnot((crop_beg[0], crop_beg[1]))
        zstart = r.zstart + crop_beg[2] * r.zinc
        # Optionally change the datatype (given in bytes per sample)
        datatype = {
            'unchanged': r.datatype,
            'int8':      SampleDataType.int8,
            'int16':     SampleDataType.int16,
            'float':     SampleDataType.float
        } [forcetype]

        singlevalue = None
        if not datarange:
            if r.datatype in (SampleDataType.int8, SampleDataType.int16):
                if r.raw_datarange[0] == r.raw_datarange[1]:
                    singlevalue = r.raw_datarange[0]
                    datarange = suggest_range(singlevalue, r.datatype)
                    # Note, if read datatype != write datatype the
                    # data range no longer is zero centric. On the
                    # other hand, the write datatype is then probably
                    # float which means it doesn't matter.

        print("Cropping: offset ", crop_beg, "size", crop_size, "of", r.size)
        #print("World corners now", world_corners, "was", r.corners)
        #print("Annot start now  ", annotstart, "was", r.annotstart)
        print("Data type", r.datatype, "->", datatype)
        with ZgyWriter(dstfilename,
                       compressor = ZgyCompressFactory("ZFP", snr = snr),
                       iocontext = SDCredentials(),
                       size = crop_size or r.size,
                       datatype = datatype if snr<=0 else SampleDataType.float,
                       datarange = datarange or r.datarange,
                       zunitdim = r.zunitdim,
                       zunitname = r.zunitname,
                       zunitfactor = r.zunitfactor,
                       hunitdim = r.hunitdim,
                       hunitname = r.hunitname,
                       hunitfactor = r.hunitfactor,
                       zstart = zstart,
                       zinc = r.zinc,
                       annotstart = r.annotstart,
                       annotinc = r.annotinc,
                       corners = world_corners) as w:
            opentime = time.time()
            if dstfilename != "/dev/null":
                if singlevalue is not None:
                    # The input file is known to contain just a single value.
                    # And the coding range is degenerate. How this is handled
                    # by the reader is not well defined. So, ignore the values
                    # that were read and just use the constant. Don't honor
                    # noiselevel int this case. That kludge doesn't really
                    # make sense for integral files anyway.
                    if True:
                        print("Writing a constant-value file, range",
                              datarange[0], "..", datarange[1],
                              "value", singlevalue)
                    w.writeconst((0,0,0), singlevalue, w.size, False)
                else:
                    copy_open_file(r, w, progress = progress1 or ProgressWithDots(), offset=crop_beg, noisefactor=noisefactor)
            else:
                read_and_discard(r, crop_size or r.size, progress = progress1 or ProgressWithDots(), offset=crop_beg)
            copytime = time.time()
            if dstfilename != "/dev/null":
                w.finalize(progress = progress2 or ProgressWithDots())
            finaltime = time.time()
    flushtime = time.time()
    if True:
        timing_report(w, flushtime - starttime)
    if True:
        print("Times: open {0:.2f} copy {1:.2f} final {2:.2f} flush {3:.2f}".format(
            opentime - starttime,
            copytime - opentime,
            finaltime - copytime,
            flushtime - finaltime))

def timing_report(writer, elapsed):
    bs = np.array(writer.bricksize, dtype=np.int64)
    size = np.array(writer.size, dtype=np.int64)
    paddedsize = ((size + bs - 1) // bs) * bs
    bandwidth = np.product(paddedsize) / elapsed # should I use size or padsize?
    bandwidth /= (1024*1024)
    print("Elapsed {0:7.2f} seconds, bandwidth {1:6.2f} MVoxel/s copying {2} {3} samples, exact {4:.0f} MVoxel, padded {5:.0f} MVoxel".format(
        elapsed, bandwidth, writer.datatype, tuple(size),
        np.product(size) / (1024*1024),
        np.product(paddedsize) / (1024*1024)))

def parseints(s):
    return tuple(map(int,s.split(",")))

def Main():
    # Can do this in a stand alone app, but not as part of a library.
    # The main problem is expressions like x = int(offset) + np.int32(size)
    # which works fine on Linux but fails occasionally on windows. The
    # reason is that offset gets changed to np.int32 on windows if it is
    # small enough; np.float64 otherwise. On Linux it is always changed
    # to np.int64. On windows the expression will overflow if and only
    # if offset is slightly less than np.int32.max.
    # Since I am adding extra tests anyway I'll cause exceptions to be
    # raised also on divide by error, underflow, etc.
    np.seterr(all='raise')

    parser = argparse.ArgumentParser(description='Copy a ZGY file.', epilog="""
    The output cube will have its data bricks sorted by lod, I, J, K
    for optimized reads from the cloud.""")
    parser.add_argument('input', help='ZGY input cube, local or sd://')
    parser.add_argument('output', help='ZGY output cube, local or sd://')
    parser.add_argument('--offset', nargs=1, default=["0,0,0"], type=str,
                        help='i,j,k Starting point in the source cube. Negative numbers count from the end.')
    parser.add_argument('--size', nargs=1, default=["0,0,0"], type=str,
                        help='i,j,k size of data to be copied. Zero means to end of cube.')
    parser.add_argument('--center', action='store_true', help="Ignore --offset, crop out center of cube.")
    parser.add_argument('--forcetype', default='unchanged', choices=['unchanged', 'int8', 'int16', 'float'], help='Produce a cube of this type instead of keeping the input type. If converting from float to int then the coding range must already be correct.')
    parser.add_argument('--datarange', nargs=2, default=None, metavar=('MIN', 'MAX'), type=float, help='Required when converting from float to integral types.')
    parser.add_argument('--noisefactor', nargs=1, default=[0], metavar='FACTOR', type=float, help='Amount of noise to add. E.g. 5000 means add 1/5000 of total data range.')
    parser.add_argument('--overwrite', action='store_true', help="Quietly overwrite the output file if it exists already.")
    parser.add_argument('--snr', default=0, type=int,
                        help='Pass 10..70 for lossy compression, 99 for lossless, 0 for uncompressed.')
    args = parser.parse_args()
    #print(args)
    args.offset = parseints(args.offset[0])
    args.size = parseints(args.size[0])
    args.datarange = tuple(args.datarange) if args.datarange is not None else None
    args.noisefactor = args.noisefactor[0]
    #print(args)
    if not args.input or not args.output:
        print("File names cannot be empty.", file=sys.stderr)
        sys.exit(1)
    if not args.overwrite and os.path.exists(args.output):
        print('Output file "{0}" already exists.'.format(args.output), file=sys.stderr)
        print('If you really meant to overwrite it, specify --overwrite.', file=sys.stderr)
        sys.exit(1)
    copy(args.input, args.output, args.offset, args.size, args.center, args.forcetype, args.datarange, args.noisefactor, args.snr)

if __name__ == "__main__":
    Main()

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



#!/usr/bin/env python3

"""
Implement a simple copy command that exercises the new code.
"""

import numpy as np
import os, sys
from ..api import ZgyReader, ZgyWriter, ProgressWithDots, SampleDataType
from ..test.utils import SDCredentials
from ..iterator import readall

def suggest_range(value, dt):
    """Special case handling for inconsistent all-constant files."""
    dt_lo, dt_hi = {SampleDataType.int8: (-128, +127),
                    SampleDataType.int16: (-32768, +32767)}[dt]
    if value == 0:
        return (dt_lo / dt_hi, 1)
    elif value > 0:
        return (0, value * (1 - dt_hi / dt_lo))
    else:
        return (value * (1 - dt_lo / dt_hi), 0)

def copy_open_file(r, w, progress):
    """Simple example of manually iterating over files."""
    def _roundup(x, step): return ((x + step - 1) // step) * step
    blocksize = (r.bricksize[0], r.bricksize[1],
                 _roundup(r.size[2], r.bricksize[2]))
    total = (((r.size[0] + blocksize[0] - 1) // blocksize[0]) *
             ((r.size[1] + blocksize[1] - 1) // blocksize[1]))
    done = 0
    data = np.zeros(blocksize, dtype=np.float32)
    for ii in range(0, r.size[0], blocksize[0]):
        for jj in range(0, r.size[1], blocksize[1]):
            r.read((ii, jj, 0), data)
            w.write((ii, jj, 0), data)
            done += 1
            if progress: progress(done, total)

def copy_open_v2(r, w, progress):
    """Fewer lines of code but more going on behind the scenes."""
    for datastart, datasize, data in readall(r, progress=progress):
        w.write(datastart, data)

def copy(srcfilename, dstfilename):
    with ZgyReader(srcfilename, iocontext = SDCredentials()) as r:
        if r.datatype in (SampleDataType.int8, SampleDataType.int16):
            if r.raw_datarange[0] == r.raw_datarange[1]:
                datarange = suggest_range(r.raw_datarange[0], r.datatype)
                with ZgyWriter(dstfilename, templatename=srcfilename,
                               datarange = datarange,
                               iocontext = SDCredentials()) as w:
                    w.writeconst((0,0,0), r.raw_datarange[0], w.size, False)
                    w.finalize(progress=ProgressWithDots())
                    return
        with ZgyWriter(dstfilename, templatename=srcfilename,
                       iocontext = SDCredentials()) as w:
            copy_open_file(r, w, progress=ProgressWithDots())
            w.finalize(progress=ProgressWithDots())

if __name__ == "__main__":
    np.seterr(all='raise')
    copy(sys.argv[1], sys.argv[2])

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
